# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""An extensible ASCII table reader and writer.

basic.py:
  Basic table read / write functionality for simple character
  delimited files with various options for column header definition.

:Copyright: Smithsonian Astrophysical Observatory (2011)
:Author: Tom Aldcroft (aldcroft@head.cfa.harvard.edu)
"""

import re

from . import core


class BasicHeader(core.BaseHeader):
    """
    Basic table Header Reader.

    Set a few defaults for common ascii table formats
    (start at line 0, comments begin with ``#`` and possibly white space)
    """

    start_line = 0
    comment = r"\s*#"
    write_comment = "# "


class BasicData(core.BaseData):
    """
    Basic table Data Reader.

    Set a few defaults for common ascii table formats
    (start at line 1, comments begin with ``#`` and possibly white space)
    """

    start_line = 1
    comment = r"\s*#"
    write_comment = "# "


class Basic(core.BaseReader):
    r"""Character-delimited table with a single header line at the top.

    Lines beginning with a comment character (default='#') as the first
    non-whitespace character are comments.

    Example table::

      # Column definition is the first uncommented line
      # Default delimiter is the space character.
      apples oranges pears

      # Data starts after the header column definition, blank lines ignored
      1 2 3
      4 5 6
    """

    _format_name = "basic"
    _description = "Basic table with custom delimiters"
    _io_registry_format_aliases = ["ascii"]

    header_class = BasicHeader
    data_class = BasicData


class NoHeaderHeader(BasicHeader):
    """
    Reader for table header without a header.

    Set the start of header line number to `None`, which tells the basic
    reader there is no header line.
    """

    start_line = None


class NoHeaderData(BasicData):
    """
    Reader for table data without a header.

    Data starts at first uncommented line since there is no header line.
    """

    start_line = 0


class NoHeader(Basic):
    """Character-delimited table with no header line.

    When reading, columns are autonamed using header.auto_format which defaults
    to "col%d".  Otherwise this reader the same as the :class:`Basic` class
    from which it is derived.  Example::

      # Table data
      1 2 "hello there"
      3 4 world

    """

    _format_name = "no_header"
    _description = "Basic table with no headers"
    header_class = NoHeaderHeader
    data_class = NoHeaderData


class CommentedHeaderHeader(BasicHeader):
    """
    Header class for which the column definition line starts with the
    comment character.  See the :class:`CommentedHeader` class  for an example.
    """

    def process_lines(self, lines):
        """
        Return only lines that start with the comment regexp.  For these
        lines strip out the matching characters.
        """
        re_comment = re.compile(self.comment)
        for line in lines:
            match = re_comment.match(line)
            if match:
                yield line[match.end() :]

    def write(self, lines):
        lines.append(self.write_comment + self.splitter.join(self.colnames))


class CommentedHeader(Basic):
    """Character-delimited table with column names in a comment line.

    When reading, ``header_start`` can be used to specify the
    line index of column names, and it can be a negative index (for example -1
    for the last commented line).  The default delimiter is the <space>
    character.

    This matches the format produced by ``np.savetxt()``, with ``delimiter=','``,
    and ``header='<comma-delimited-column-names-list>'``.

    Example::

      # col1 col2 col3
      # Comment line
      1 2 3
      4 5 6

    """

    _format_name = "commented_header"
    _description = "Column names in a commented line"

    header_class = CommentedHeaderHeader
    data_class = NoHeaderData

    def read(self, table):
        """
        Read input data (file-like object, filename, list of strings, or
        single string) into a Table and return the result.
        """
        out = super().read(table)

        # Strip off the comment line set as the header line for
        # commented_header format (first by default).
        if "comments" in out.meta:
            idx = self.header.start_line
            if idx < 0:
                idx = len(out.meta["comments"]) + idx
            out.meta["comments"] = (
                out.meta["comments"][:idx] + out.meta["comments"][idx + 1 :]
            )
            if not out.meta["comments"]:
                del out.meta["comments"]

        return out

    def write_header(self, lines, meta):
        """
        Write comment lines after, rather than before, the header.
        """
        self.header.write(lines)
        self.header.write_comments(lines, meta)


class TabHeaderSplitter(core.DefaultSplitter):
    """Split lines on tab and do not remove whitespace."""

    delimiter = "\t"

    def process_line(self, line):
        return line + "\n"


class TabDataSplitter(TabHeaderSplitter):
    """
    Don't strip data value whitespace since that is significant in TSV tables.
    """

    process_val = None
    skipinitialspace = False


class TabHeader(BasicHeader):
    """
    Reader for header of tables with tab separated header.
    """

    splitter_class = TabHeaderSplitter


class TabData(BasicData):
    """
    Reader for data of tables with tab separated data.
    """

    splitter_class = TabDataSplitter


class Tab(Basic):
    """Tab-separated table.

    Unlike the :class:`Basic` reader, whitespace is not stripped from the
    beginning and end of either lines or individual column values.

    Example::

      col1 <tab> col2 <tab> col3
      # Comment line
      1 <tab> 2 <tab> 5

    """

    _format_name = "tab"
    _description = "Basic table with tab-separated values"
    header_class = TabHeader
    data_class = TabData


class CsvSplitter(core.DefaultSplitter):
    """
    Split on comma for CSV (comma-separated-value) tables.
    """

    delimiter = ","


class CsvHeader(BasicHeader):
    """
    Header that uses the :class:`astropy.io.ascii.basic.CsvSplitter`.
    """

    splitter_class = CsvSplitter
    comment = None
    write_comment = None


class CsvData(BasicData):
    """
    Data that uses the :class:`astropy.io.ascii.basic.CsvSplitter`.
    """

    splitter_class = CsvSplitter
    fill_values = [(core.masked, "")]
    comment = None
    write_comment = None


class Csv(Basic):
    """CSV (comma-separated-values) table.

    This file format may contain rows with fewer entries than the number of
    columns, a situation that occurs in output from some spreadsheet editors.
    The missing entries are marked as masked in the output table.

    Masked values (indicated by an empty '' field value when reading) are
    written out in the same way with an empty ('') field.  This is different
    from the typical default for `astropy.io.ascii` in which missing values are
    indicated by ``--``.

    Since the `CSV format <https://tools.ietf.org/html/rfc4180>`_ does not
    formally support comments, any comments defined for the table via
    ``tbl.meta['comments']`` are ignored by default. If you would still like to
    write those comments then include a keyword ``comment='#'`` to the
    ``write()`` call.

    Example::

      num,ra,dec,radius,mag
      1,32.23222,10.1211
      2,38.12321,-88.1321,2.2,17.0

    """

    _format_name = "csv"
    _io_registry_format_aliases = ["csv"]
    _io_registry_can_write = True
    _io_registry_suffix = ".csv"
    _description = "Comma-separated-values"

    header_class = CsvHeader
    data_class = CsvData

    def inconsistent_handler(self, str_vals, ncols):
        """
        Adjust row if it is too short.

        If a data row is shorter than the header, add empty values to make it the
        right length.
        Note that this will *not* be called if the row already matches the header.

        Parameters
        ----------
        str_vals : list
            A list of value strings from the current row of the table.
        ncols : int
            The expected number of entries from the table header.

        Returns
        -------
        str_vals : list
            List of strings to be parsed into data entries in the output table.
        """
        if len(str_vals) < ncols:
            str_vals.extend((ncols - len(str_vals)) * [""])

        return str_vals


class RdbHeader(TabHeader):
    """
    Header for RDB tables.
    """

    col_type_map = {"n": core.NumType, "s": core.StrType}

    def get_type_map_key(self, col):
        return col.raw_type[-1]

    def get_cols(self, lines):
        """
        Initialize the header Column objects from the table ``lines``.

        This is a specialized get_cols for the RDB type:
        Line 0: RDB col names
        Line 1: RDB col definitions
        Line 2+: RDB data rows

        Parameters
        ----------
        lines : list
            List of table lines

        Returns
        -------
        None

        """
        header_lines = self.process_lines(lines)  # this is a generator
        header_vals_list = [hl for _, hl in zip(range(2), self.splitter(header_lines))]
        if len(header_vals_list) != 2:
            raise ValueError("RDB header requires 2 lines")
        self.names, raw_types = header_vals_list

        if len(self.names) != len(raw_types):
            raise core.InconsistentTableError(
                "RDB header mismatch between number of column names and column types."
            )

        if any(not re.match(r"\d*(N|S)$", x, re.IGNORECASE) for x in raw_types):
            raise core.InconsistentTableError(
                f"RDB types definitions do not all match [num](N|S): {raw_types}"
            )

        self._set_cols_from_names()
        for col, raw_type in zip(self.cols, raw_types):
            col.raw_type = raw_type
            col.type = self.get_col_type(col)

    def write(self, lines):
        lines.append(self.splitter.join(self.colnames))
        rdb_types = []
        for col in self.cols:
            # Check if dtype.kind is string or unicode.  See help(np.core.numerictypes)
            rdb_type = "S" if col.info.dtype.kind in ("S", "U") else "N"
            rdb_types.append(rdb_type)

        lines.append(self.splitter.join(rdb_types))


class RdbData(TabData):
    """
    Data reader for RDB data. Starts reading at line 2.
    """

    start_line = 2


class Rdb(Tab):
    """Tab-separated file with an extra line after the column definition line that
    specifies either numeric (N) or string (S) data.

    See: https://www.drdobbs.com/rdb-a-unix-command-line-database/199101326

    Example::

      col1 <tab> col2 <tab> col3
      N <tab> S <tab> N
      1 <tab> 2 <tab> 5

    """

    _format_name = "rdb"
    _io_registry_format_aliases = ["rdb"]
    _io_registry_suffix = ".rdb"
    _description = "Tab-separated with a type definition header line"

    header_class = RdbHeader
    data_class = RdbData



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import matplotlib.lines
import matplotlib.pyplot as plt
import matplotlib.text
import numpy as np
import pytest
from matplotlib import rc_context
from matplotlib.figure import Figure
from matplotlib.patches import Circle, Rectangle

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.tests.figures import figure_test
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning
from astropy.visualization.wcsaxes import WCSAxes, add_beam, add_scalebar
from astropy.visualization.wcsaxes.frame import EllipticalFrame
from astropy.visualization.wcsaxes.patches import Quadrangle, SphericalCircle
from astropy.wcs import WCS


class BaseImageTests:
    @classmethod
    def setup_class(cls):
        msx_header = get_pkg_data_filename("data/msx_header")
        cls.msx_header = fits.Header.fromtextfile(msx_header)

        rosat_header = get_pkg_data_filename("data/rosat_header")
        cls.rosat_header = fits.Header.fromtextfile(rosat_header)

        twoMASS_k_header = get_pkg_data_filename("data/2MASS_k_header")
        cls.twoMASS_k_header = fits.Header.fromtextfile(twoMASS_k_header)

        cube_header = get_pkg_data_filename("data/cube_header")
        cls.cube_header = fits.Header.fromtextfile(cube_header)

        slice_header = get_pkg_data_filename("data/slice_header")
        cls.slice_header = fits.Header.fromtextfile(slice_header)

    def teardown_method(self, method):
        plt.close("all")


class TestBasic(BaseImageTests):
    @figure_test
    def test_tight_layout(self):
        # Check that tight_layout works on a WCSAxes.
        fig = plt.figure(figsize=(8, 6))
        for i in (1, 2):
            fig.add_subplot(2, 1, i, projection=WCS(self.msx_header))
        fig.tight_layout()
        return fig

    @figure_test
    def test_image_plot(self):
        # Test for plotting image and also setting values of ticks
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8], projection=WCS(self.msx_header), aspect="equal"
        )
        ax.set_xlim(-0.5, 148.5)
        ax.set_ylim(-0.5, 148.5)
        ax.coords[0].set_ticks([-0.30, 0.0, 0.20] * u.degree, size=5, width=1)
        return fig

    @figure_test
    def test_axes_off(self):
        # Test for turning the axes off
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection=WCS(self.msx_header))
        ax.imshow(np.arange(12).reshape((3, 4)))
        ax.set_axis_off()
        return fig

    @figure_test
    @pytest.mark.parametrize("axisbelow", [True, False, "line"])
    def test_axisbelow(self, axisbelow):
        # Test that tick marks, labels, and gridlines are drawn with the
        # correct zorder controlled by the axisbelow property.
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8], projection=WCS(self.msx_header), aspect="equal"
        )
        ax.set_axisbelow(axisbelow)
        ax.set_xlim(-0.5, 148.5)
        ax.set_ylim(-0.5, 148.5)
        ax.coords[0].set_ticks([-0.30, 0.0, 0.20] * u.degree, size=5, width=1)
        ax.grid()
        ax.coords[0].set_auto_axislabel(False)
        ax.coords[1].set_auto_axislabel(False)

        # Add an image (default zorder=0).
        ax.imshow(np.zeros((64, 64)))

        # Add a patch (default zorder=1).
        r = Rectangle((30.0, 50.0), 60.0, 50.0, facecolor="green", edgecolor="red")
        ax.add_patch(r)

        # Add a line (default zorder=2).
        ax.plot([32, 128], [32, 128], linewidth=10)

        return fig

    @figure_test
    def test_contour_overlay(self):
        # Test for overlaying contours on images
        path = get_pkg_data_filename("galactic_center/gc_msx_e.fits")
        with fits.open(path) as pf:
            data = pf[0].data

        wcs_msx = WCS(self.msx_header)

        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.15, 0.15, 0.8, 0.8],
            projection=WCS(self.twoMASS_k_header),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 720.5)
        ax.set_ylim(-0.5, 720.5)

        # Overplot contour
        ax.contour(
            data,
            transform=ax.get_transform(wcs_msx),
            colors="orange",
            levels=[2.5e-5, 5e-5, 1.0e-4],
        )
        ax.coords[0].set_ticks(size=5, width=1)
        ax.coords[1].set_ticks(size=5, width=1)
        ax.set_xlim(0.0, 720.0)
        ax.set_ylim(0.0, 720.0)

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_contourf_overlay(self):
        # Test for overlaying contours on images
        path = get_pkg_data_filename("galactic_center/gc_msx_e.fits")
        with fits.open(path) as pf:
            data = pf[0].data

        wcs_msx = WCS(self.msx_header)

        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.15, 0.15, 0.8, 0.8],
            projection=WCS(self.twoMASS_k_header),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 720.5)
        ax.set_ylim(-0.5, 720.5)

        # Overplot contour
        ax.contourf(
            data, transform=ax.get_transform(wcs_msx), levels=[2.5e-5, 5e-5, 1.0e-4]
        )
        ax.coords[0].set_ticks(size=5, width=1)
        ax.coords[1].set_ticks(size=5, width=1)
        ax.set_xlim(0.0, 720.0)
        ax.set_ylim(0.0, 720.0)

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_overlay_features_image(self):
        # Test for overlaying grid, changing format of ticks, setting spacing
        # and number of ticks

        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.25, 0.25, 0.65, 0.65], projection=WCS(self.msx_header), aspect="equal"
        )

        # Change the format of the ticks
        ax.coords[0].set_major_formatter("dd:mm:ss")
        ax.coords[1].set_major_formatter("dd:mm:ss.ssss")

        # Overlay grid on image
        ax.grid(color="red", alpha=1.0, lw=1, linestyle="dashed")

        # Set the spacing of ticks on the 'glon' axis to 4 arcsec
        ax.coords["glon"].set_ticks(spacing=4 * u.arcsec, size=5, width=1)

        # Set the number of ticks on the 'glat' axis to 9
        ax.coords["glat"].set_ticks(number=9, size=5, width=1)

        # Set labels on axes
        ax.coords["glon"].set_axislabel("Galactic Longitude", minpad=1.6)
        ax.coords["glat"].set_axislabel("Galactic Latitude", minpad=-0.75)

        # Change the frame linewidth and color
        ax.coords.frame.set_color("red")
        ax.coords.frame.set_linewidth(2)

        assert ax.coords.frame.get_color() == "red"
        assert ax.coords.frame.get_linewidth() == 2

        return fig

    @figure_test
    def test_curvilinear_grid_patches_image(self):
        # Overlay curvilinear grid and patches on image

        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8], projection=WCS(self.rosat_header), aspect="equal"
        )

        ax.set_xlim(-0.5, 479.5)
        ax.set_ylim(-0.5, 239.5)

        ax.grid(color="black", alpha=1.0, lw=1, linestyle="dashed")

        p = Circle((300, 100), radius=40, ec="yellow", fc="none")
        ax.add_patch(p)

        p = Circle(
            (30.0, 20.0),
            radius=20.0,
            ec="orange",
            fc="none",
            transform=ax.get_transform("world"),
        )
        ax.add_patch(p)

        p = Circle(
            (60.0, 50.0),
            radius=20.0,
            ec="red",
            fc="none",
            transform=ax.get_transform("fk5"),
        )
        ax.add_patch(p)

        p = Circle(
            (40.0, 60.0),
            radius=20.0,
            ec="green",
            fc="none",
            transform=ax.get_transform("galactic"),
        )
        ax.add_patch(p)

        return fig

    @figure_test
    def test_cube_slice_image(self):
        # Test for cube slicing

        fig = plt.figure()
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8],
            projection=WCS(self.cube_header),
            slices=(50, "y", "x"),
            aspect="equal",
        )

        ax.set_xlim(-0.5, 52.5)
        ax.set_ylim(-0.5, 106.5)

        ax.coords[2].set_axislabel("Velocity m/s")

        ax.coords[1].set_ticks(spacing=0.2 * u.deg, width=1)
        ax.coords[2].set_ticks(spacing=400 * u.m / u.s, width=1)

        ax.coords[1].set_ticklabel(exclude_overlapping=True)
        ax.coords[2].set_ticklabel(exclude_overlapping=True)

        ax.coords[0].grid(grid_type="contours", color="purple", linestyle="solid")
        ax.coords[1].grid(grid_type="contours", color="orange", linestyle="solid")
        ax.coords[2].grid(grid_type="contours", color="red", linestyle="solid")

        return fig

    @figure_test
    def test_cube_slice_image_lonlat(self):
        # Test for cube slicing. Here we test with longitude and latitude since
        # there is some longitude-specific code in _update_grid_contour.

        fig = plt.figure()
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8],
            projection=WCS(self.cube_header),
            slices=("x", "y", 50),
            aspect="equal",
        )

        ax.set_xlim(-0.5, 106.5)
        ax.set_ylim(-0.5, 106.5)

        ax.coords[0].grid(grid_type="contours", color="blue", linestyle="solid")
        ax.coords[1].grid(grid_type="contours", color="red", linestyle="solid")

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_plot_coord(self):
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.15, 0.15, 0.8, 0.8],
            projection=WCS(self.twoMASS_k_header),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 720.5)
        ax.set_ylim(-0.5, 720.5)

        c = SkyCoord(266 * u.deg, -29 * u.deg)
        lines = ax.plot_coord(c, "o")

        # Test that plot_coord returns the results from ax.plot
        assert isinstance(lines, list)
        assert isinstance(lines[0], matplotlib.lines.Line2D)

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_scatter_coord(self):
        from matplotlib.collections import PathCollection

        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.15, 0.15, 0.8, 0.8],
            projection=WCS(self.twoMASS_k_header),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 720.5)
        ax.set_ylim(-0.5, 720.5)

        c = SkyCoord(266 * u.deg, -29 * u.deg)
        sc = ax.scatter_coord(c, marker="o")

        # Test that plot_coord returns the results from ax.plot
        assert isinstance(sc, PathCollection)

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_text_coord(self):
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.15, 0.15, 0.8, 0.8],
            projection=WCS(self.twoMASS_k_header),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 720.5)
        ax.set_ylim(-0.5, 720.5)

        c = SkyCoord(266 * u.deg, -29 * u.deg)
        text = ax.text_coord(c, "Sample Label", color="blue", ha="right", va="top")

        # Test that plot_coord returns the results from ax.text
        assert isinstance(text, matplotlib.text.Text)

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_plot_line(self):
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.15, 0.15, 0.8, 0.8],
            projection=WCS(self.twoMASS_k_header),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 720.5)
        ax.set_ylim(-0.5, 720.5)

        c = SkyCoord([266, 266.8] * u.deg, [-29, -28.9] * u.deg)
        ax.plot_coord(c)

        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)

        return fig

    @figure_test
    def test_changed_axis_units(self):
        # Test to see if changing the units of axis works
        fig = plt.figure()
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8],
            projection=WCS(self.cube_header),
            slices=(50, "y", "x"),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 52.5)
        ax.set_ylim(-0.5, 106.5)
        ax.coords[0].set_ticks_position("")
        ax.coords[0].set_ticklabel_position("")
        ax.coords[0].set_axislabel_position("")
        ax.coords[1].set_ticks_position("lr")
        ax.coords[1].set_ticklabel_position("l")
        ax.coords[1].set_axislabel_position("l")
        ax.coords[2].set_ticks_position("bt")
        ax.coords[2].set_ticklabel_position("b")
        ax.coords[2].set_axislabel_position("b")
        ax.coords[2].set_major_formatter("x.xx")
        ax.coords[2].set_format_unit(u.km / u.s)
        ax.coords[2].set_axislabel("Velocity km/s")
        ax.coords[1].set_ticks(width=1)
        ax.coords[2].set_ticks(width=1)
        ax.coords[1].set_ticklabel(exclude_overlapping=True)
        ax.coords[2].set_ticklabel(exclude_overlapping=True)

        return fig

    @figure_test
    def test_minor_ticks(self):
        # Test for drawing minor ticks
        fig = plt.figure()
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8],
            projection=WCS(self.cube_header),
            slices=(50, "y", "x"),
            aspect="equal",
        )
        ax.set_xlim(-0.5, 52.5)
        ax.set_ylim(-0.5, 106.5)
        ax.coords[0].set_ticks_position("")
        ax.coords[0].set_ticklabel_position("")
        ax.coords[0].set_axislabel_position("")
        ax.coords[1].set_ticks_position("lr")
        ax.coords[1].set_ticklabel_position("l")
        ax.coords[1].set_axislabel_position("l")
        ax.coords[2].set_ticks_position("bt")
        ax.coords[2].set_ticklabel_position("b")
        ax.coords[2].set_axislabel_position("b")
        ax.coords[2].set_ticklabel(exclude_overlapping=True)
        ax.coords[1].set_ticklabel(exclude_overlapping=True)
        ax.coords[2].display_minor_ticks(True)
        ax.coords[1].display_minor_ticks(True)
        ax.coords[2].set_minor_frequency(3)
        ax.coords[1].set_minor_frequency(10)

        return fig

    @figure_test
    def test_ticks_labels(self):
        fig = plt.figure(figsize=(6, 6))
        ax = WCSAxes(fig, [0.1, 0.1, 0.7, 0.7], wcs=None)
        fig.add_axes(ax)
        ax.set_xlim(-0.5, 2)
        ax.set_ylim(-0.5, 2)
        ax.coords[0].set_ticks(size=10, color="blue", alpha=0.2, width=1)
        ax.coords[1].set_ticks(size=20, color="red", alpha=0.9, width=1)
        ax.coords[0].set_ticks_position("all")
        ax.coords[1].set_ticks_position("all")
        ax.coords[0].set_axislabel("X-axis", size=20)
        ax.coords[1].set_axislabel(
            "Y-axis",
            color="green",
            size=25,
            weight="regular",
            style="normal",
            family="cmtt10",
        )
        ax.coords[0].set_axislabel_position("t")
        ax.coords[1].set_axislabel_position("r")
        ax.coords[0].set_ticklabel(
            color="purple",
            size=15,
            alpha=1,
            weight="light",
            style="normal",
            family="cmss10",
        )
        ax.coords[1].set_ticklabel(
            color="black", size=18, alpha=0.9, weight="bold", family="cmr10"
        )
        ax.coords[0].set_ticklabel_position("all")
        ax.coords[1].set_ticklabel_position("r")

        return fig

    @figure_test
    def test_no_ticks(self):
        # Check that setting no ticks works
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_axes(
            [0.1, 0.1, 0.8, 0.8], projection=WCS(self.msx_header), aspect="equal"
        )
        ax.set_xlim(-0.5, 148.5)
        ax.set_ylim(-0.5, 148.5)
        ax.coords[0].set_ticks(number=0)
        ax.coords[0].grid(True)
        return fig

    @figure_test
    def test_rcparams(self):
        # Test custom rcParams

        with rc_context(
            {
                "axes.labelcolor": "purple",
                "axes.labelsize": 14,
                "axes.labelweight": "bold",
                "axes.linewidth": 3,
                "axes.facecolor": "0.5",
                "axes.edgecolor": "green",
                "xtick.color": "red",
                "xtick.labelsize": 8,
                "xtick.direction": "in",
                "xtick.minor.visible": True,
                "xtick.minor.size": 5,
                "xtick.major.size": 20,
                "xtick.major.width": 3,
                "xtick.major.pad": 10,
                "grid.color": "blue",
                "grid.linestyle": ":",
                "grid.linewidth": 1,
                "grid.alpha": 0.5,
            }
        ):
            fig = plt.figure(figsize=(6, 6))
            ax = WCSAxes(fig, [0.15, 0.1, 0.7, 0.7], wcs=None)
            fig.add_axes(ax)
            ax.set_xlim(-0.5, 2)
            ax.set_ylim(-0.5, 2)
            ax.grid()
            ax.set_xlabel("X label")
            ax.set_ylabel("Y label")
            ax.coords[0].set_ticklabel(exclude_overlapping=True)
            ax.coords[1].set_ticklabel(exclude_overlapping=True)
            return fig

    @figure_test
    def test_tick_angles(self):
        # Test that tick marks point in the correct direction, even when the
        # axes limits extend only over a few FITS pixels. Addresses #45, #46.
        w = WCS()
        w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        w.wcs.crval = [90, 70]
        w.wcs.cdelt = [16, 16]
        w.wcs.crpix = [1, 1]
        w.wcs.radesys = "ICRS"
        w.wcs.equinox = 2000.0
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection=w)
        ax.set_xlim(1, -1)
        ax.set_ylim(-1, 1)
        ax.grid(color="gray", alpha=0.5, linestyle="solid")
        ax.coords["ra"].set_ticks(color="red", size=20)
        ax.coords["dec"].set_ticks(color="red", size=20)
        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)
        return fig

    @figure_test
    def test_tick_angles_non_square_axes(self):
        # Test that tick marks point in the correct direction, even when the
        # axes limits extend only over a few FITS pixels, and the axes are
        # non-square.
        w = WCS()
        w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        w.wcs.crval = [90, 70]
        w.wcs.cdelt = [16, 16]
        w.wcs.crpix = [1, 1]
        w.wcs.radesys = "ICRS"
        w.wcs.equinox = 2000.0
        fig = plt.figure(figsize=(6, 3))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection=w)
        ax.set_xlim(1, -1)
        ax.set_ylim(-1, 1)
        ax.grid(color="gray", alpha=0.5, linestyle="solid")
        ax.coords["ra"].set_ticks(color="red", size=20)
        ax.coords["dec"].set_ticks(color="red", size=20)
        # In previous versions, all angle axes defaulted to being displayed in
        # degrees. We now automatically show RA axes in hour angle units, but
        # for backward-compatibility with previous reference images we
        # explicitly use degrees here.
        ax.coords[0].set_format_unit(u.degree)
        return fig

    @figure_test
    def test_set_coord_type(self):
        # Test for setting coord_type
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes(
            [0.2, 0.2, 0.6, 0.6], projection=WCS(self.msx_header), aspect="equal"
        )
        ax.set_xlim(-0.5, 148.5)
        ax.set_ylim(-0.5, 148.5)
        ax.coords[0].set_coord_type("scalar")
        ax.coords[1].set_coord_type("scalar")
        ax.coords[0].set_major_formatter("x.xxx")
        ax.coords[1].set_major_formatter("x.xxx")
        ax.coords[0].set_ticklabel(exclude_overlapping=True)
        ax.coords[1].set_ticklabel(exclude_overlapping=True)
        return fig

    @figure_test
    def test_ticks_regression(self):
        # Regression test for a bug that caused ticks aligned exactly with a
        # sampled frame point to not appear. This also checks that tick labels
        # don't get added more than once, and that no error occurs when e.g.
        # the top part of the frame is all at the same coordinate as one of the
        # potential ticks (which causes the tick angle calculation to return
        # NaN).
        wcs = WCS(self.slice_header)
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes([0.25, 0.25, 0.5, 0.5], projection=wcs, aspect="auto")
        limits = wcs.wcs_world2pix([0, 0], [35e3, 80e3], 0)[1]
        ax.set_ylim(*limits)
        ax.coords[0].set_ticks(spacing=0.002 * u.deg)
        ax.coords[1].set_ticks(spacing=5 * u.km / u.s)
        ax.coords[0].set_ticklabel(alpha=0.5)  # to see multiple labels
        ax.coords[1].set_ticklabel(alpha=0.5)
        ax.coords[0].set_ticklabel_position("all")
        ax.coords[1].set_ticklabel_position("all")
        return fig

    @figure_test
    def test_axislabels_regression(self):
        # Regression test for a bug that meant that if tick labels were made
        # invisible with ``set_visible(False)``, they were still added to the
        # list of bounding boxes for tick labels, but with default values of 0
        # to 1, which caused issues.
        wcs = WCS(self.msx_header)
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes([0.25, 0.25, 0.5, 0.5], projection=wcs, aspect="auto")
        ax.coords[0].set_axislabel("Label 1")
        ax.coords[1].set_axislabel("Label 2")
        ax.coords[1].set_axislabel_visibility_rule("always")
        ax.coords[1].ticklabels.set_visible(False)
        return fig

    @figure_test(savefig_kwargs={"bbox_inches": "tight"})
    def test_noncelestial_angular(self, tmp_path):
        # Regression test for a bug that meant that when passing a WCS that had
        # angular axes and using set_coord_type to set the coordinates to
        # longitude/latitude, but where the WCS wasn't recognized as celestial,
        # the WCS units are not converted to deg, so we can't assume that
        # transform will always return degrees.

        wcs = WCS(naxis=2)

        wcs.wcs.ctype = ["solar-x", "solar-y"]
        wcs.wcs.cunit = ["arcsec", "arcsec"]

        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_subplot(1, 1, 1, projection=wcs)

        ax.imshow(np.zeros([1024, 1024]), origin="lower")

        ax.coords[0].set_coord_type("longitude", coord_wrap=180 * u.deg)
        ax.coords[1].set_coord_type("latitude")

        ax.coords[0].set_major_formatter("s.s")
        ax.coords[1].set_major_formatter("s.s")

        ax.coords[0].set_format_unit(u.arcsec, show_decimal_unit=False)
        ax.coords[1].set_format_unit(u.arcsec, show_decimal_unit=False)

        ax.grid(color="white", ls="solid")

        # Force drawing (needed for format_coord)
        fig.savefig(tmp_path / "nothing")

        assert ax.format_coord(512, 512) == "513.0 513.0 (world)"

        return fig

    @figure_test
    def test_patches_distortion(self, tmp_path):
        # Check how patches get distorted (and make sure that scatter markers
        # and SphericalCircle don't)

        wcs = WCS(self.msx_header)
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes([0.25, 0.25, 0.5, 0.5], projection=wcs, aspect="equal")

        # Pixel coordinates
        r = Rectangle((30.0, 50.0), 60.0, 50.0, edgecolor="green", facecolor="none")
        ax.add_patch(r)

        # FK5 coordinates
        r = Rectangle(
            (266.4, -28.9),
            0.3,
            0.3,
            edgecolor="cyan",
            facecolor="none",
            transform=ax.get_transform("fk5"),
        )
        ax.add_patch(r)

        # FK5 coordinates
        c = Circle(
            (266.4, -29.1),
            0.15,
            edgecolor="magenta",
            facecolor="none",
            transform=ax.get_transform("fk5"),
        )
        ax.add_patch(c)

        # Pixel coordinates
        ax.scatter(
            [40, 100, 130],
            [30, 130, 60],
            s=100,
            edgecolor="red",
            facecolor=(1, 0, 0, 0.5),
        )

        # World coordinates (should not be distorted)
        ax.scatter(
            266.78238,
            -28.769255,
            transform=ax.get_transform("fk5"),
            s=300,
            edgecolor="red",
            facecolor="none",
        )

        # World coordinates (should not be distorted)
        r1 = SphericalCircle(
            (266.4 * u.deg, -29.1 * u.deg),
            0.15 * u.degree,
            edgecolor="purple",
            facecolor="none",
            transform=ax.get_transform("fk5"),
        )
        ax.add_patch(r1)

        r2 = SphericalCircle(
            SkyCoord(266.4 * u.deg, -29.1 * u.deg),
            0.15 * u.degree,
            edgecolor="purple",
            facecolor="none",
            transform=ax.get_transform("fk5"),
        )

        with pytest.warns(
            AstropyUserWarning,
            match="Received `center` of representation type "
            "<class 'astropy.coordinates.*CartesianRepresentation'> "
            "will be converted to SphericalRepresentation",
        ):
            r3 = SphericalCircle(
                SkyCoord(
                    x=-0.05486461,
                    y=-0.87204803,
                    z=-0.48633538,
                    representation_type="cartesian",
                ),
                0.15 * u.degree,
                edgecolor="purple",
                facecolor="none",
                transform=ax.get_transform("fk5"),
            )

        ax.coords[0].set_ticklabel_visible(False)
        ax.coords[1].set_ticklabel_visible(False)

        # Test to verify that SphericalCircle works irrespective of whether
        # the input(center) is a tuple or a SkyCoord object.
        assert (r1.get_xy() == r2.get_xy()).all()
        assert np.allclose(r1.get_xy(), r3.get_xy())
        assert np.allclose(r2.get_xy()[0], [266.4, -29.25])

        return fig

    @figure_test
    def test_quadrangle(self, tmp_path):
        # Test that Quadrangle can have curved edges while Rectangle does not
        wcs = WCS(self.msx_header)
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes([0.25, 0.25, 0.5, 0.5], projection=wcs, aspect="equal")
        ax.set_xlim(0, 10000)
        ax.set_ylim(-10000, 0)

        # Add a quadrangle patch (100 degrees by 20 degrees)
        q = Quadrangle(
            (255, -70) * u.deg,
            100 * u.deg,
            20 * u.deg,
            label="Quadrangle",
            edgecolor="blue",
            facecolor="none",
            transform=ax.get_transform("icrs"),
        )
        ax.add_patch(q)

        # Add a rectangle patch (100 degrees by 20 degrees)
        r = Rectangle(
            (255, -70),
            100,
            20,
            label="Rectangle",
            edgecolor="red",
            facecolor="none",
            linestyle="--",
            transform=ax.get_transform("icrs"),
        )
        ax.add_patch(r)

        ax.coords[0].set_ticklabel_visible(False)
        ax.coords[1].set_ticklabel_visible(False)

        return fig

    @figure_test
    def test_beam_shape_from_args(self, tmp_path):
        # Test for adding the beam shape with the beam parameters as arguments
        wcs = WCS(self.msx_header)
        fig = plt.figure(figsize=(4, 3))
        ax = fig.add_axes([0.2, 0.2, 0.6, 0.6], projection=wcs, aspect="equal")
        ax.set_xlim(-10, 10)
        ax.set_ylim(-10, 10)

        add_beam(
            ax,
            major=2 * u.arcmin,
            minor=1 * u.arcmin,
            angle=-30 * u.degree,
            corner="bottom right",
            frame=True,
            borderpad=0.0,
            pad=1.0,
            color="black",
        )

        return fig

    @figure_test
    def test_beam_shape_from_header(self, tmp_path):
        # Test for adding the beam shape with the beam parameters from a header
        hdr = self.msx_header
        hdr["BMAJ"] = (2 * u.arcmin).to(u.degree).value
        hdr["BMIN"] = (1 * u.arcmin).to(u.degree).value
        hdr["BPA"] = 30.0

        wcs = WCS(hdr)
        fig = plt.figure(figsize=(4, 3))
        ax = fig.add_axes([0.2, 0.2, 0.6, 0.6], projection=wcs, aspect="equal")
        ax.set_xlim(-10, 10)
        ax.set_ylim(-10, 10)

        add_beam(ax, header=hdr)

        return fig

    @figure_test
    def test_scalebar(self, tmp_path):
        # Test for adding a scale bar
        wcs = WCS(self.msx_header)
        fig = plt.figure(figsize=(4, 3))
        ax = fig.add_axes([0.2, 0.2, 0.6, 0.6], projection=wcs, aspect="equal")
        ax.set_xlim(-10, 10)
        ax.set_ylim(-10, 10)

        add_scalebar(
            ax,
            2 * u.arcmin,
            label="2'",
            corner="top right",
            borderpad=1.0,
            label_top=True,
        )

        return fig

    @figure_test
    def test_elliptical_frame(self):
        # Regression test for a bug (astropy/astropy#6063) that caused labels to
        # be incorrectly simplified.

        wcs = WCS(self.msx_header)
        fig = plt.figure(figsize=(5, 3))
        fig.add_axes([0.2, 0.2, 0.6, 0.6], projection=wcs, frame_class=EllipticalFrame)
        return fig

    @figure_test
    def test_hms_labels(self):
        # This tests the appearance of the hms superscripts in tick labels
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes(
            [0.3, 0.2, 0.65, 0.6], projection=WCS(self.twoMASS_k_header), aspect="equal"
        )
        ax.set_xlim(-0.5, 0.5)
        ax.set_ylim(-0.5, 0.5)
        ax.coords[0].set_ticks(spacing=0.2 * 15 * u.arcsec)
        return fig

    @figure_test(style={"text.usetex": True})
    def test_latex_labels(self):
        fig = plt.figure(figsize=(3, 3))
        ax = fig.add_axes(
            [0.3, 0.2, 0.65, 0.6], projection=WCS(self.twoMASS_k_header), aspect="equal"
        )
        ax.set_xlim(-0.5, 0.5)
        ax.set_ylim(-0.5, 0.5)
        ax.coords[0].set_ticks(spacing=0.2 * 15 * u.arcsec)
        return fig

    @figure_test
    def test_tick_params(self):
        # This is a test to make sure that tick_params works correctly. We try
        # and test as much as possible with a single reference image.

        wcs = WCS()
        wcs.wcs.ctype = ["lon", "lat"]

        fig = plt.figure(figsize=(6, 6))

        # The first subplot tests:
        # - that plt.tick_params works
        # - that by default both axes are changed
        # - changing the tick direction and appearance, the label appearance and padding
        ax = fig.add_subplot(2, 2, 1, projection=wcs)
        plt.tick_params(
            direction="in",
            length=20,
            width=5,
            pad=6,
            labelsize=6,
            color="red",
            labelcolor="blue",
        )

        ax.coords[0].set_auto_axislabel(False)
        ax.coords[1].set_auto_axislabel(False)

        # The second subplot tests:
        # - that specifying grid parameters doesn't actually cause the grid to
        #   be shown (as expected)
        # - that axis= can be given integer coordinates or their string name
        # - that the tick positioning works (bottom/left/top/right)
        # Make sure that we can pass things that can index coords
        ax = fig.add_subplot(2, 2, 2, projection=wcs)
        plt.tick_params(
            axis=0,
            direction="in",
            length=20,
            width=5,
            pad=4,
            labelsize=6,
            color="red",
            labelcolor="blue",
            bottom=True,
            grid_color="purple",
        )
        plt.tick_params(
            axis="lat",
            direction="out",
            labelsize=8,
            color="blue",
            labelcolor="purple",
            left=True,
            right=True,
            grid_color="red",
        )

        ax.coords[0].set_auto_axislabel(False)
        ax.coords[1].set_auto_axislabel(False)

        # The third subplot tests:
        # - that ax.tick_params works
        # - that the grid has the correct settings once shown explicitly
        # - that we can use axis='x' and axis='y'
        ax = fig.add_subplot(2, 2, 3, projection=wcs)
        ax.tick_params(
            axis="x",
            direction="in",
            length=20,
            width=5,
            pad=20,
            labelsize=6,
            color="red",
            labelcolor="blue",
            bottom=True,
            grid_color="purple",
        )
        ax.tick_params(
            axis="y",
            direction="out",
            labelsize=8,
            color="blue",
            labelcolor="purple",
            left=True,
            right=True,
            grid_color="red",
        )
        plt.grid()

        ax.coords[0].set_auto_axislabel(False)
        ax.coords[1].set_auto_axislabel(False)

        # The final subplot tests:
        # - that we can use tick_params on a specific coordinate
        # - that the label positioning can be customized
        # - that the colors argument works
        # - that which='minor' works
        ax = fig.add_subplot(2, 2, 4, projection=wcs)
        ax.coords[0].tick_params(
            length=4,
            pad=2,
            colors="orange",
            labelbottom=True,
            labeltop=True,
            labelsize=10,
        )
        ax.coords[1].display_minor_ticks(True)
        ax.coords[1].tick_params(which="minor", length=6)

        ax.coords[0].set_auto_axislabel(False)
        ax.coords[1].set_auto_axislabel(False)

        return fig


@pytest.fixture
def wave_wcs_1d():
    wcs = WCS(naxis=1)
    wcs.wcs.ctype = ["WAVE"]
    wcs.wcs.cunit = ["m"]
    wcs.wcs.crpix = [1]
    wcs.wcs.cdelt = [5]
    wcs.wcs.crval = [45]
    wcs.wcs.set()
    return wcs


@figure_test
def test_1d_plot_1d_wcs(wave_wcs_1d):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=wave_wcs_1d)
    (lines,) = ax.plot([10, 12, 14, 12, 10])

    ax.set_xlabel("this is the x-axis")
    ax.set_ylabel("this is the y-axis")

    return fig


@figure_test
def test_1d_plot_1d_wcs_format_unit(wave_wcs_1d):
    """
    This test ensures that the format unit is updated and displayed for both
    the axis ticks and default axis labels.
    """
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=wave_wcs_1d)
    (lines,) = ax.plot([10, 12, 14, 12, 10])

    ax.coords[0].set_format_unit("nm")

    return fig


@pytest.fixture
def spatial_wcs_2d():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ["GLON-TAN", "GLAT-TAN"]
    wcs.wcs.crpix = [3.0] * 2
    wcs.wcs.cdelt = [15] * 2
    wcs.wcs.crval = [50.0] * 2
    wcs.wcs.set()
    return wcs


@figure_test
def test_1d_plot_2d_wcs_correlated(spatial_wcs_2d):
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=spatial_wcs_2d, slices=("x", 0))
    (lines,) = ax.plot([10, 12, 14, 12, 10], "-o", color="orange")

    ax.coords["glon"].set_ticks(color="red")
    ax.coords["glon"].set_ticklabel(color="red")
    ax.coords["glon"].grid(color="red")

    ax.coords["glat"].set_ticks(color="blue")
    ax.coords["glat"].set_ticklabel(color="blue")
    ax.coords["glat"].grid(color="blue")

    return fig


@pytest.fixture
def spatial_wcs_2d_small_angle():
    """
    This WCS has an almost linear correlation between the pixel and world axes
    close to the reference pixel.
    """
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ["HPLN-TAN", "HPLT-TAN"]
    wcs.wcs.crpix = [3.0] * 2
    wcs.wcs.cdelt = [10 / 3600, 5 / 3600]
    wcs.wcs.crval = [0] * 2
    wcs.wcs.set()
    return wcs


@pytest.mark.parametrize(
    "slices, bottom_axis",
    [
        # Remember SLLWCS takes slices in array order
        (np.s_[0, :], "custom:pos.helioprojective.lon"),
        (np.s_[:, 0], "custom:pos.helioprojective.lat"),
    ],
)
@figure_test
def test_1d_plot_1d_sliced_low_level_wcs(
    spatial_wcs_2d_small_angle, slices, bottom_axis
):
    """
    Test that a SLLWCS through a coupled 2D WCS plots as line OK.
    """
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=spatial_wcs_2d_small_angle[slices])
    (lines,) = ax.plot([10, 12, 14, 12, 10], "-o", color="orange")

    # Draw to trigger rendering the ticks.
    plt.draw()

    assert ax.coords[bottom_axis].ticks.get_visible_axes() == ["b"]

    return fig


@pytest.mark.parametrize(
    "slices, bottom_axis", [(("x", 0), "hpln"), ((0, "x"), "hplt")]
)
@figure_test
def test_1d_plot_put_varying_axis_on_bottom_lon(
    spatial_wcs_2d_small_angle, slices, bottom_axis
):
    """
    When we plot a 1D slice through spatial axes, we want to put the axis which
    actually changes on the bottom.

    For example an aligned wcs, pixel grid where you plot a lon slice through a
    lat axis, you would end up with no ticks on the bottom as the lon doesn't
    change, and a set of lat ticks on the top because it does but it's the
    correlated axis not the actual one you are plotting against.
    """
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection=spatial_wcs_2d_small_angle, slices=slices)
    (lines,) = ax.plot([10, 12, 14, 12, 10], "-o", color="orange")

    # Draw to trigger rendering the ticks.
    plt.draw()

    assert ax.coords[bottom_axis].ticks.get_visible_axes() == ["b"]

    return fig


@figure_test
def test_allsky_labels_wrap():
    # Regression test for a bug that caused some tick labels to not be shown
    # when looking at all-sky maps in the case where coord_wrap < 360

    fig = plt.figure(figsize=(4, 4))

    icen = 0

    for ctype in [("GLON-CAR", "GLAT-CAR"), ("HGLN-CAR", "HGLT-CAR")]:
        for cen in [0, 90, 180, 270]:
            icen += 1

            wcs = WCS(naxis=2)
            wcs.wcs.ctype = ctype
            wcs.wcs.crval = cen, 0
            wcs.wcs.crpix = 360.5, 180.5
            wcs.wcs.cdelt = -0.5, 0.5

            ax = fig.add_subplot(8, 1, icen, projection=wcs)
            ax.set_xlim(-0.5, 719.5)
            ax.coords[0].set_ticks(spacing=50 * u.deg)
            ax.coords[0].set_ticks_position("b")
            ax.coords[0].set_auto_axislabel(False)
            ax.coords[1].set_auto_axislabel(False)
            ax.coords[1].set_ticklabel_visible(False)
            ax.coords[1].set_ticks_visible(False)

    fig.subplots_adjust(hspace=2, left=0.05, right=0.95, bottom=0.1, top=0.95)

    return fig


@figure_test
def test_tickable_gridlines():
    wcs = WCS(
        {
            "naxis": 2,
            "naxis1": 360,
            "naxis2": 180,
            "crpix1": 180.5,
            "crpix2": 90.5,
            "cdelt1": -1,
            "cdelt2": 1,
            "ctype1": "RA---CAR",
            "ctype2": "DEC--CAR",
        }
    )

    fig = Figure()
    ax = fig.add_subplot(projection=wcs)
    ax.set_xlim(-0.5, 360 - 0.5)
    ax.set_ylim(-0.5, 150 - 0.5)

    lon, lat = ax.coords
    lon.grid()
    lat.grid()

    overlay = ax.get_coords_overlay("galactic")
    overlay[0].set_ticks(spacing=30 * u.deg)
    overlay[1].set_ticks(spacing=30 * u.deg)

    # Test both single-character and multi-character names
    overlay[1].add_tickable_gridline("g", -30 * u.deg)
    overlay[0].add_tickable_gridline("const-glon", 30 * u.deg)

    overlay[0].grid(color="magenta")
    overlay[0].set_ticklabel_position("gt")
    overlay[0].set_ticklabel(color="magenta")
    overlay[0].set_axislabel("Galactic longitude", color="magenta")

    overlay[1].grid(color="blue")
    overlay[1].set_ticklabel_position(("const-glon", "r"))
    overlay[1].set_ticklabel(color="blue")
    overlay[1].set_axislabel("Galactic latitude", color="blue")

    return fig



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module provides utility functions for the models package.
"""

import warnings

# pylint: disable=invalid-name
from collections import UserDict
from inspect import signature

import numpy as np

from astropy import units as u
from astropy.utils.decorators import deprecated

__all__ = ["poly_map_domain", "comb", "ellipse_extent"]


def make_binary_operator_eval(oper, f, g):
    """
    Given a binary operator (as a callable of two arguments) ``oper`` and
    two callables ``f`` and ``g`` which accept the same arguments,
    returns a *new* function that takes the same arguments as ``f`` and ``g``,
    but passes the outputs of ``f`` and ``g`` in the given ``oper``.

    ``f`` and ``g`` are assumed to return tuples (which may be 1-tuples).  The
    given operator is applied element-wise to tuple outputs).

    Example
    -------
    >>> from operator import add
    >>> def prod(x, y):
    ...     return (x * y,)
    ...
    >>> sum_of_prod = make_binary_operator_eval(add, prod, prod)
    >>> sum_of_prod(3, 5)
    (30,)
    """
    return lambda inputs, params: tuple(
        oper(x, y) for x, y in zip(f(inputs, params), g(inputs, params))
    )


def poly_map_domain(oldx, domain, window):
    """
    Map domain into window by shifting and scaling.

    Parameters
    ----------
    oldx : array
          original coordinates
    domain : list or tuple of length 2
          function domain
    window : list or tuple of length 2
          range into which to map the domain
    """
    domain = np.array(domain, dtype=np.float64)
    window = np.array(window, dtype=np.float64)
    if domain.shape != (2,) or window.shape != (2,):
        raise ValueError('Expected "domain" and "window" to be a tuple of size 2.')
    scl = (window[1] - window[0]) / (domain[1] - domain[0])
    off = (window[0] * domain[1] - window[1] * domain[0]) / (domain[1] - domain[0])
    return off + scl * oldx


def _validate_domain_window(value):
    if value is not None:
        if np.asanyarray(value).shape != (2,):
            raise ValueError("domain and window should be tuples of size 2.")
        return tuple(value)
    return value


@deprecated("5.3", alternative="math.comb")
def comb(N, k):
    """
    The number of combinations of N things taken k at a time.

    Parameters
    ----------
    N : int, array
        Number of things.
    k : int, array
        Number of elements taken.

    """
    if (k > N) or (N < 0) or (k < 0):
        return 0
    val = 1
    for j in range(min(k, N - k)):
        val = (val * (N - j)) / (j + 1)
    return val


def array_repr_oneline(array):
    """
    Represents a multi-dimensional Numpy array flattened onto a single line.
    """
    r = np.array2string(array, separator=", ", suppress_small=True)
    return " ".join(line.strip() for line in r.splitlines())


def combine_labels(left, right):
    """
    For use with the join operator &: Combine left input/output labels with
    right input/output labels.

    If none of the labels conflict then this just returns a sum of tuples.
    However if *any* of the labels conflict, this appends '0' to the left-hand
    labels and '1' to the right-hand labels so there is no ambiguity).
    """
    if set(left).intersection(right):
        left = tuple(label + "0" for label in left)
        right = tuple(label + "1" for label in right)

    return left + right


def ellipse_extent(a, b, theta):
    """
    Calculates the half size of a box encapsulating a rotated 2D
    ellipse.

    Parameters
    ----------
    a : float or `~astropy.units.Quantity`
        The ellipse semimajor axis.
    b : float or `~astropy.units.Quantity`
        The ellipse semiminor axis.
    theta : float or `~astropy.units.Quantity` ['angle']
        The rotation angle as an angular quantity
        (`~astropy.units.Quantity` or `~astropy.coordinates.Angle`) or
        a value in radians (as a float). The rotation angle increases
        counterclockwise.

    Returns
    -------
    offsets : tuple
        The absolute value of the offset distances from the ellipse center that
        define its bounding box region, ``(dx, dy)``.

    Examples
    --------
    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        from astropy.modeling.models import Ellipse2D
        from astropy.modeling.utils import ellipse_extent, render_model

        amplitude = 1
        x0 = 50
        y0 = 50
        a = 30
        b = 10
        theta = np.pi / 4

        model = Ellipse2D(amplitude, x0, y0, a, b, theta)
        dx, dy = ellipse_extent(a, b, theta)
        limits = [x0 - dx, x0 + dx, y0 - dy, y0 + dy]
        model.bounding_box = limits

        image = render_model(model)

        plt.imshow(image, cmap='binary', interpolation='nearest', alpha=.5,
                  extent = limits)
        plt.show()
    """
    from .parameters import Parameter  # prevent circular import

    if isinstance(theta, Parameter):
        if theta.quantity is None:
            theta = theta.value
        else:
            theta = theta.quantity

    t = np.arctan2(-b * np.tan(theta), a)
    dx = a * np.cos(t) * np.cos(theta) - b * np.sin(t) * np.sin(theta)

    t = np.arctan2(b, a * np.tan(theta))
    dy = b * np.sin(t) * np.cos(theta) + a * np.cos(t) * np.sin(theta)

    if isinstance(dx, u.Quantity) or isinstance(dy, u.Quantity):
        return np.abs(u.Quantity([dx, dy], subok=True))
    return np.abs([dx, dy])


def get_inputs_and_params(func):
    """
    Given a callable, determine the input variables and the
    parameters.

    Parameters
    ----------
    func : callable

    Returns
    -------
    inputs, params : tuple
        Each entry is a list of inspect.Parameter objects
    """
    sig = signature(func)

    inputs = []
    params = []
    for param in sig.parameters.values():
        if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            raise ValueError("Signature must not have *args or **kwargs")
        if param.default == param.empty:
            inputs.append(param)
        else:
            params.append(param)

    return inputs, params


def _combine_equivalency_dict(keys, eq1=None, eq2=None):
    # Given two dictionaries that give equivalencies for a set of keys, for
    # example input value names, return a dictionary that includes all the
    # equivalencies
    eq = {}
    for key in keys:
        eq[key] = []
        if eq1 is not None and key in eq1:
            eq[key].extend(eq1[key])
        if eq2 is not None and key in eq2:
            eq[key].extend(eq2[key])
    return eq


def _to_radian(value):
    """Convert ``value`` to radian."""
    if isinstance(value, u.Quantity):
        return value.to(u.rad)
    return np.deg2rad(value)


def _to_orig_unit(value, raw_unit=None, orig_unit=None):
    """Convert value with ``raw_unit`` to ``orig_unit``."""
    if raw_unit is not None:
        return (value * raw_unit).to(orig_unit)
    return np.rad2deg(value)


class _ConstraintsDict(UserDict):
    """
    Wrapper around UserDict to allow updating the constraints
    on a Parameter when the dictionary is updated.
    """

    def __init__(self, model, constraint_type):
        self._model = model
        self.constraint_type = constraint_type
        c = {}
        for name in model.param_names:
            param = getattr(model, name)
            c[name] = getattr(param, constraint_type)
        super().__init__(c)

    def __setitem__(self, key, val):
        super().__setitem__(key, val)
        param = getattr(self._model, key)
        setattr(param, self.constraint_type, val)


class _SpecialOperatorsDict(UserDict):
    """
    Wrapper around UserDict to allow for better tracking of the Special
    Operators for CompoundModels. This dictionary is structured so that
    one cannot inadvertently overwrite an existing special operator.

    Parameters
    ----------
    unique_id: int
        the last used unique_id for a SPECIAL OPERATOR
    special_operators: dict
        a dictionary containing the special_operators

    Notes
    -----
    Direct setting of operators (`dict[key] = value`) into the
    dictionary has been deprecated in favor of the `.add(name, value)`
    method, so that unique dictionary keys can be generated and tracked
    consistently.
    """

    def __init__(self, unique_id=0, special_operators={}):
        super().__init__(special_operators)
        self._unique_id = unique_id

    def _set_value(self, key, val):
        if key in self:
            raise ValueError(f'Special operator "{key}" already exists')
        else:
            super().__setitem__(key, val)

    def __setitem__(self, key, val):
        self._set_value(key, val)
        warnings.warn(
            DeprecationWarning(
                """
            Special operator dictionary assignment has been deprecated.
            Please use `.add` instead, so that you can capture a unique
            key for your operator.
            """
            )
        )

    def _get_unique_id(self):
        self._unique_id += 1

        return self._unique_id

    def add(self, operator_name, operator):
        """
        Adds a special operator to the dictionary, and then returns the
        unique key that the operator is stored under for later reference.

        Parameters
        ----------
        operator_name: str
            the name for the operator
        operator: function
            the actual operator function which will be used

        Returns
        -------
        the unique operator key for the dictionary
            `(operator_name, unique_id)`
        """
        key = (operator_name, self._get_unique_id())

        self._set_value(key, operator)

        return key



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Handles the "VOUnit" unit format.
"""

import copy
import keyword
import re
import warnings

from . import core, generic, utils


class VOUnit(generic.Generic):
    """
    The IVOA standard for units used by the VO.

    This is an implementation of `Units in the VO 1.0
    <http://www.ivoa.net/documents/VOUnits/>`_.
    """

    _explicit_custom_unit_regex = re.compile(r"^[YZEPTGMkhdcmunpfazy]?'((?!\d)\w)+'$")
    _custom_unit_regex = re.compile(r"^((?!\d)\w)+$")
    _custom_units = {}
    _space = "."
    _scale_unit_separator = ""

    @staticmethod
    def _generate_unit_names():
        from astropy import units as u
        from astropy.units import required_by_vounit as uvo

        names = {}
        deprecated_names = set()
        # The tropical year is missing here compared to the standard
        bases = [
            "A", "a", "adu", "arcmin", "arcsec", "barn", "beam", "bin",
            "C", "cd", "chan", "count", "ct", "d", "D", "deg", "erg", "eV",
            "F", "g", "G", "H", "h", "Hz", "J", "Jy", "K", "lm", "lx", "lyr",
            "m", "mag", "min", "mol", "N", "Ohm", "Pa", "pc", "ph", "photon",
            "pix", "pixel", "R", "rad", "Ry", "s", "S", "solLum", "solMass",
            "solRad", "sr", "T", "u", "V", "voxel", "W", "Wb", "yr",
        ]  # fmt: skip
        binary_bases = ["bit", "byte", "B"]
        simple_units = ["Angstrom", "angstrom", "AU", "au", "Ba", "dB", "mas", "Sun"]
        si_prefixes = [
            "y", "z", "a", "f", "p", "n", "u", "m", "c", "d",
            "", "da", "h", "k", "M", "G", "T", "P", "E", "Z", "Y"
        ]  # fmt: skip
        # While zebi and yobi are part of the standard for binary prefixes,
        # they are not implemented here due to computation limitations
        binary_prefixes = ["Ki", "Mi", "Gi", "Ti", "Pi", "Ei"]
        deprecated_units = {
            "angstrom", "Angstrom", "Ba", "barn", "erg", "G", "ta",
        }  # fmt: skip

        def do_defines(bases, prefixes, skips=[]):
            for base in bases:
                for prefix in prefixes:
                    key = prefix + base
                    if key in skips:
                        continue
                    if keyword.iskeyword(key):
                        continue

                    names[key] = getattr(u if hasattr(u, key) else uvo, key)
                    if base in deprecated_units:
                        deprecated_names.add(key)

        do_defines(bases, si_prefixes, ["pct", "pcount", "yd"])
        do_defines(binary_bases, si_prefixes + binary_prefixes, ["dB", "dbyte"])
        do_defines(simple_units, [""])

        return names, deprecated_names, []

    @classmethod
    def parse(cls, s, debug=False):
        if s in ("unknown", "UNKNOWN"):
            return None
        if s == "":
            return core.dimensionless_unscaled
        # Check for excess solidi, but exclude fractional exponents (allowed)
        if s.count("/") > 1 and s.count("/") - len(re.findall(r"\(\d+/\d+\)", s)) > 1:
            raise core.UnitsError(
                f"'{s}' contains multiple slashes, which is "
                "disallowed by the VOUnit standard."
            )
        result = cls._do_parse(s, debug=debug)
        if hasattr(result, "function_unit"):
            raise ValueError("Function units are not yet supported in VOUnit.")
        return result

    @classmethod
    def _get_unit(cls, t):
        try:
            return super()._get_unit(t)
        except ValueError:
            if cls._explicit_custom_unit_regex.match(t.value):
                return cls._def_custom_unit(t.value)

            if cls._custom_unit_regex.match(t.value):
                warnings.warn(
                    f"Unit {t.value!r} not supported by the VOUnit standard. "
                    + utils.did_you_mean_units(
                        t.value,
                        cls._units,
                        cls._deprecated_units,
                        cls._to_decomposed_alternative,
                    ),
                    core.UnitsWarning,
                )

                return cls._def_custom_unit(t.value)

            raise

    @classmethod
    def _parse_unit(cls, unit, detailed_exception=True):
        if unit not in cls._units:
            raise ValueError()

        if unit in cls._deprecated_units:
            utils.unit_deprecation_warning(
                unit, cls._units[unit], "VOUnit", cls._to_decomposed_alternative
            )

        return cls._units[unit]

    @classmethod
    def _get_unit_name(cls, unit):
        # The da- and d- prefixes are discouraged.  This has the
        # effect of adding a scale to value in the result.
        if isinstance(unit, core.PrefixUnit):
            if unit._represents.scale == 10.0:
                raise ValueError(
                    f"In '{unit}': VOUnit can not represent units with the 'da' "
                    "(deka) prefix"
                )
            elif unit._represents.scale == 0.1:
                raise ValueError(
                    f"In '{unit}': VOUnit can not represent units with the 'd' "
                    "(deci) prefix"
                )

        name = super()._get_unit_name(unit)

        if unit in cls._custom_units.values():
            return name

        if name not in cls._units:
            raise ValueError(f"Unit {name!r} is not part of the VOUnit standard")

        if name in cls._deprecated_units:
            utils.unit_deprecation_warning(
                name, unit, "VOUnit", cls._to_decomposed_alternative
            )

        return name

    @classmethod
    def _def_custom_unit(cls, unit):
        def def_base(name):
            if name in cls._custom_units:
                return cls._custom_units[name]

            if name.startswith("'"):
                return core.def_unit(
                    [name[1:-1], name],
                    format={"vounit": name},
                    namespace=cls._custom_units,
                )
            else:
                return core.def_unit(name, namespace=cls._custom_units)

        if unit in cls._custom_units:
            return cls._custom_units[unit]

        for short, full, factor in core.si_prefixes:
            for prefix in short:
                if unit.startswith(prefix):
                    base_name = unit[len(prefix) :]
                    base_unit = def_base(base_name)
                    return core.PrefixUnit(
                        [prefix + x for x in base_unit.names],
                        core.CompositeUnit(
                            factor, [base_unit], [1], _error_check=False
                        ),
                        format={"vounit": prefix + base_unit.names[-1]},
                        namespace=cls._custom_units,
                    )

        return def_base(unit)

    @classmethod
    def _format_superscript(cls, number):
        return f"**({number})" if "/" in number or "." in number else f"**{number}"

    @classmethod
    def format_exponential_notation(cls, val, format_spec=".8g"):
        return super().format_exponential_notation(val, format_spec)

    @classmethod
    def _format_fraction(cls, scale, numerator, denominator, *, fraction="inline"):
        if not (fraction is True or fraction == "inline"):
            raise ValueError(
                "format {cls.name!r} only supports inline fractions,"
                f"not fraction={fraction!r}."
            )

        if cls._space in denominator:
            denominator = f"({denominator})"
        if scale and numerator == "1":
            return f"{scale}/{denominator}"
        return f"{scale}{numerator}/{denominator}"

    @classmethod
    def to_string(cls, unit, fraction=False):
        from astropy.units import core

        # Remove units that aren't known to the format
        unit = utils.decompose_to_known_units(unit, cls._get_unit_name)

        if unit.physical_type == "dimensionless" and unit.scale != 1:
            raise core.UnitScaleError(
                "The VOUnit format is not able to "
                "represent scale for dimensionless units. "
                f"Multiply your data by {unit.scale:e}."
            )

        return super().to_string(unit, fraction=fraction)

    @classmethod
    def _to_decomposed_alternative(cls, unit):
        from astropy.units import core

        try:
            s = cls.to_string(unit)
        except core.UnitScaleError:
            scale = unit.scale
            unit = copy.copy(unit)
            unit._scale = 1.0
            return f"{cls.to_string(unit)} (with data multiplied by {scale})"
        return s



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy import units as u
from astropy.coordinates import representation as r
from astropy.coordinates.attributes import (
    EarthLocationAttribute,
    QuantityAttribute,
    TimeAttribute,
)
from astropy.coordinates.baseframe import (
    BaseCoordinateFrame,
    RepresentationMapping,
    base_doc,
)
from astropy.utils.decorators import format_doc

__all__ = ["HADec"]


doc_components = """
    ha : `~astropy.coordinates.Angle`, optional, keyword-only
        The Hour Angle for this object (``dec`` must also be given and
        ``representation`` must be None).
    dec : `~astropy.coordinates.Angle`, optional, keyword-only
        The Declination for this object (``ha`` must also be given and
        ``representation`` must be None).
    distance : `~astropy.units.Quantity` ['length'], optional, keyword-only
        The Distance for this object along the line-of-sight.

    pm_ha_cosdec : `~astropy.units.Quantity` ['angular speed'], optional, keyword-only
        The proper motion in hour angle (including the ``cos(dec)`` factor) for
        this object (``pm_dec`` must also be given).
    pm_dec : `~astropy.units.Quantity` ['angular speed'], optional, keyword-only
        The proper motion in declination for this object (``pm_ha_cosdec`` must
        also be given).
    radial_velocity : `~astropy.units.Quantity` ['speed'], optional, keyword-only
        The radial velocity of this object."""

doc_footer = """
    Other parameters
    ----------------
    obstime : `~astropy.time.Time`
        The time at which the observation is taken.  Used for determining the
        position and orientation of the Earth.
    location : `~astropy.coordinates.EarthLocation`
        The location on the Earth.  This can be specified either as an
        `~astropy.coordinates.EarthLocation` object or as anything that can be
        transformed to an `~astropy.coordinates.ITRS` frame.
    pressure : `~astropy.units.Quantity` ['pressure']
        The atmospheric pressure as an `~astropy.units.Quantity` with pressure
        units.  This is necessary for performing refraction corrections.
        Setting this to 0 (the default) will disable refraction calculations
        when transforming to/from this frame.
    temperature : `~astropy.units.Quantity` ['temperature']
        The ground-level temperature as an `~astropy.units.Quantity` in
        deg C.  This is necessary for performing refraction corrections.
    relative_humidity : `~astropy.units.Quantity` ['dimensionless'] or number.
        The relative humidity as a dimensionless quantity between 0 to 1.
        This is necessary for performing refraction corrections.
    obswl : `~astropy.units.Quantity` ['length']
        The average wavelength of observations as an `~astropy.units.Quantity`
         with length units.  This is necessary for performing refraction
         corrections.

    Notes
    -----
    The refraction model is based on that implemented in ERFA, which is fast
    but becomes inaccurate for altitudes below about 5 degrees.  Near and below
    altitudes of 0, it can even give meaningless answers, and in this case
    transforming to HADec and back to another frame can give highly discrepant
    results.  For much better numerical stability, leave the ``pressure`` at
    ``0`` (the default), thereby disabling the refraction correction and
    yielding "topocentric" equatorial coordinates.
    """


@format_doc(base_doc, components=doc_components, footer=doc_footer)
class HADec(BaseCoordinateFrame):
    """
    A coordinate or frame in the Hour Angle-Declination system (Equatorial
    coordinates) with respect to the WGS84 ellipsoid.  Hour Angle is oriented
    with respect to upper culmination such that the hour angle is negative to
    the East and positive to the West.

    This frame is assumed to *include* refraction effects if the ``pressure``
    frame attribute is non-zero.

    The frame attributes are listed under **Other Parameters**, which are
    necessary for transforming from HADec to some other system.
    """

    frame_specific_representation_info = {
        r.SphericalRepresentation: [
            RepresentationMapping("lon", "ha", u.hourangle),
            RepresentationMapping("lat", "dec"),
        ]
    }

    default_representation = r.SphericalRepresentation
    default_differential = r.SphericalCosLatDifferential

    obstime = TimeAttribute(
        default=None, doc="The reference time (e.g., time of observation)"
    )
    location = EarthLocationAttribute(
        default=None, doc="The location on Earth of the observer"
    )
    pressure = QuantityAttribute(default=0, unit=u.hPa, doc="The atmospheric pressure")
    temperature = QuantityAttribute(
        default=0, unit=u.deg_C, doc="The ground-level temperature"
    )
    relative_humidity = QuantityAttribute(
        default=0, unit=u.dimensionless_unscaled, doc="The relative humidity"
    )
    obswl = QuantityAttribute(
        default=1 * u.micron,
        unit=u.micron,
        doc="The average wavelength of observations",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.has_data:
            self._set_data_lon_wrap_angle(self.data)

    @staticmethod
    def _set_data_lon_wrap_angle(data):
        if hasattr(data, "lon"):
            data.lon.wrap_angle = 180.0 * u.deg
        return data

    def represent_as(self, base, s="base", in_frame_units=False):
        """
        Ensure the wrap angle for any spherical
        representations.
        """
        data = super().represent_as(base, s, in_frame_units=in_frame_units)
        self._set_data_lon_wrap_angle(data)
        return data


# self-transform defined in icrs_observed_transforms.py



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from io import StringIO

import numpy as np

import astropy.units as u
from astropy.io import ascii
from astropy.table import QTable

from .common import assert_almost_equal, assert_equal


def assert_equal_splitlines(arg1, arg2):
    assert_equal(arg1.splitlines(), arg2.splitlines())


def test_read_normal():
    """Normal SimpleRST Table"""
    table = """
# comment (with blank line above)
======= =========
   Col1      Col2
======= =========
   1.2    "hello"
   2.4  's worlds
======= =========
"""
    reader = ascii.get_reader(reader_cls=ascii.RST)
    dat = reader.read(table)
    assert_equal(dat.colnames, ["Col1", "Col2"])
    assert_almost_equal(dat[1][0], 2.4)
    assert_equal(dat[0][1], '"hello"')
    assert_equal(dat[1][1], "'s worlds")


def test_read_normal_names():
    """Normal SimpleRST Table with provided column names"""
    table = """
# comment (with blank line above)
======= =========
   Col1      Col2
======= =========
   1.2    "hello"
   2.4  's worlds
======= =========
"""
    reader = ascii.get_reader(reader_cls=ascii.RST, names=("name1", "name2"))
    dat = reader.read(table)
    assert_equal(dat.colnames, ["name1", "name2"])
    assert_almost_equal(dat[1][0], 2.4)


def test_read_normal_names_include():
    """Normal SimpleRST Table with provided column names"""
    table = """
# comment (with blank line above)
=======  ========== ======
   Col1     Col2      Col3
=======  ========== ======
   1.2     "hello"       3
   2.4    's worlds      7
=======  ========== ======
"""
    reader = ascii.get_reader(
        reader_cls=ascii.RST,
        names=("name1", "name2", "name3"),
        include_names=("name1", "name3"),
    )
    dat = reader.read(table)
    assert_equal(dat.colnames, ["name1", "name3"])
    assert_almost_equal(dat[1][0], 2.4)
    assert_equal(dat[0][1], 3)


def test_read_normal_exclude():
    """Nice, typical SimpleRST table with col name excluded"""
    table = """
======= ==========
  Col1     Col2
======= ==========
  1.2     "hello"
  2.4    's worlds
======= ==========
"""
    reader = ascii.get_reader(reader_cls=ascii.RST, exclude_names=("Col1",))
    dat = reader.read(table)
    assert_equal(dat.colnames, ["Col2"])
    assert_equal(dat[1][0], "'s worlds")


def test_read_unbounded_right_column():
    """The right hand column should be allowed to overflow"""
    table = """
# comment (with blank line above)
===== ===== ====
 Col1  Col2 Col3
===== ===== ====
 1.2    2    Hello
 2.4     4   Worlds
===== ===== ====
"""
    reader = ascii.get_reader(reader_cls=ascii.RST)
    dat = reader.read(table)
    assert_equal(dat[0][2], "Hello")
    assert_equal(dat[1][2], "Worlds")


def test_read_unbounded_right_column_header():
    """The right hand column should be allowed to overflow"""
    table = """
# comment (with blank line above)
===== ===== ====
 Col1  Col2 Col3Long
===== ===== ====
 1.2    2    Hello
 2.4     4   Worlds
===== ===== ====
"""
    reader = ascii.get_reader(reader_cls=ascii.RST)
    dat = reader.read(table)
    assert_equal(dat.colnames[-1], "Col3Long")


def test_read_right_indented_table():
    """We should be able to read right indented tables correctly"""
    table = """
# comment (with blank line above)
   ==== ==== ====
   Col1 Col2 Col3
   ==== ==== ====
    3    3.4  foo
    1    4.5  bar
   ==== ==== ====
"""
    reader = ascii.get_reader(reader_cls=ascii.RST)
    dat = reader.read(table)
    assert_equal(dat.colnames, ["Col1", "Col2", "Col3"])
    assert_equal(dat[0][2], "foo")
    assert_equal(dat[1][0], 1)


def test_trailing_spaces_in_row_definition():
    """Trailing spaces in the row definition column shouldn't matter"""
    table = (
        "\n"
        "# comment (with blank line above)\n"
        "   ==== ==== ====    \n"
        "   Col1 Col2 Col3\n"
        "   ==== ==== ====  \n"
        "    3    3.4  foo\n"
        "    1    4.5  bar\n"
        "   ==== ==== ====  \n"
    )
    # make sure no one accidentally deletes the trailing whitespaces in the
    # table.
    assert len(table) == 151

    reader = ascii.get_reader(reader_cls=ascii.RST)
    dat = reader.read(table)
    assert_equal(dat.colnames, ["Col1", "Col2", "Col3"])
    assert_equal(dat[0][2], "foo")
    assert_equal(dat[1][0], 1)


table = """\
====== =========== ============ ===========
  Col1    Col2        Col3        Col4
====== =========== ============ ===========
  1.2    "hello"      1           a
  2.4   's worlds          2           2
====== =========== ============ ===========
"""
dat = ascii.read(table, format="rst")


def test_write_normal():
    """Write a table as a normal SimpleRST Table"""
    out = StringIO()
    ascii.write(dat, out, format="rst")
    assert_equal_splitlines(
        out.getvalue(),
        """\
==== ========= ==== ====
Col1      Col2 Col3 Col4
==== ========= ==== ====
 1.2   "hello"    1    a
 2.4 's worlds    2    2
==== ========= ==== ====
""",
    )


def test_rst_with_header_rows():
    """Round-trip a table with header_rows specified"""
    lines = [
        "======= ======== ====",
        "   wave response ints",
        "     nm       ct     ",
        "float64  float32 int8",
        "======= ======== ====",
        "  350.0      1.0    1",
        "  950.0      2.0    2",
        "======= ======== ====",
    ]
    tbl = QTable.read(lines, format="ascii.rst", header_rows=["name", "unit", "dtype"])
    assert tbl["wave"].unit == u.nm
    assert tbl["response"].unit == u.ct
    assert tbl["wave"].dtype == np.float64
    assert tbl["response"].dtype == np.float32
    assert tbl["ints"].dtype == np.int8

    out = StringIO()
    tbl.write(out, format="ascii.rst", header_rows=["name", "unit", "dtype"])
    assert out.getvalue().splitlines() == lines



from contextlib import nullcontext

import numpy as np
import pytest
from numpy.testing import assert_allclose

import astropy.units as u
from astropy import time
from astropy.constants import c
from astropy.coordinates import (
    FK5,
    GCRS,
    ICRS,
    CartesianDifferential,
    CartesianRepresentation,
    EarthLocation,
    Galactic,
    SkyCoord,
    SpectralQuantity,
    get_body_barycentric_posvel,
)
from astropy.coordinates.sites import get_builtin_sites
from astropy.coordinates.spectral_coordinate import (
    NoDistanceWarning,
    NoVelocityWarning,
    SpectralCoord,
    _apply_relativistic_doppler_shift,
)
from astropy.table import Table
from astropy.tests.helper import (
    PYTEST_LT_8_0,
    assert_quantity_allclose,
    quantity_allclose,
)
from astropy.utils import iers
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning, AstropyWarning
from astropy.wcs.wcsapi.fitswcs import VELOCITY_FRAMES as FITSWCS_VELOCITY_FRAMES

GREENWICH = get_builtin_sites()["greenwich"]


def assert_frame_allclose(
    frame1,
    frame2,
    pos_rtol=1e-7,
    pos_atol=1 * u.m,
    vel_rtol=1e-7,
    vel_atol=1 * u.mm / u.s,
):
    # checks that:
    # - the positions are equal to within some tolerance (the relative tolerance
    #   should be dimensionless, the absolute tolerance should be a distance).
    #   note that these are the tolerances *in 3d*
    # - either both or nether frame has velocities, or if one has no velocities
    #   the other one can have zero velocities
    # - if velocities are present, they are equal to some tolerance
    # Ideally this should accept both frames and SkyCoords
    if hasattr(frame1, "frame"):  # SkyCoord-like
        frame1 = frame1.frame
    if hasattr(frame2, "frame"):  # SkyCoord-like
        frame2 = frame2.frame

    # assert (frame1.data.differentials and frame2.data.differentials or
    #         (not frame1.data.differentials and not frame2.data.differentials))
    assert frame1.is_equivalent_frame(frame2)

    frame2_in_1 = frame2.transform_to(frame1)

    assert_quantity_allclose(
        0 * u.m, frame1.separation_3d(frame2_in_1), rtol=pos_rtol, atol=pos_atol
    )

    if frame1.data.differentials:
        d1 = frame1.data.represent_as(
            CartesianRepresentation, CartesianDifferential
        ).differentials["s"]
        d2 = frame2_in_1.data.represent_as(
            CartesianRepresentation, CartesianDifferential
        ).differentials["s"]

        assert_quantity_allclose(d1.norm(d1), d1.norm(d2), rtol=vel_rtol, atol=vel_atol)


# GENERAL TESTS

# We first run through a series of cases to test different ways of initializing
# the observer and target for SpectralCoord, including for example frames,
# SkyCoords, and making sure that SpectralCoord is not sensitive to the actual
# frame or representation class.

# Local Standard of Rest
LSRD = Galactic(
    u=0.1 * u.km,
    v=0.1 * u.km,
    w=0.1 * u.km,
    U=9 * u.km / u.s,
    V=12 * u.km / u.s,
    W=7 * u.km / u.s,
    representation_type="cartesian",
    differential_type="cartesian",
)

LSRD_EQUIV = [
    LSRD,
    SkyCoord(LSRD),  # as a SkyCoord
    LSRD.transform_to(ICRS()),  # different frame
    LSRD.transform_to(ICRS()).transform_to(Galactic()),  # different representation
]


@pytest.fixture(params=[None] + LSRD_EQUIV)
def observer(request):
    return request.param


# Target located in direction of motion of LSRD with no velocities
LSRD_DIR_STATIONARY = Galactic(
    u=9 * u.km, v=12 * u.km, w=7 * u.km, representation_type="cartesian"
)

LSRD_DIR_STATIONARY_EQUIV = [
    LSRD_DIR_STATIONARY,
    SkyCoord(LSRD_DIR_STATIONARY),  # as a SkyCoord
    LSRD_DIR_STATIONARY.transform_to(FK5()),  # different frame
    # different representation
    LSRD_DIR_STATIONARY.transform_to(ICRS()).transform_to(Galactic()),
]


@pytest.fixture(params=[None] + LSRD_DIR_STATIONARY_EQUIV)
def target(request):
    return request.param


def test_create_spectral_coord_observer_target(observer, target):
    with (
        nullcontext()
        if target is None
        else pytest.warns(AstropyUserWarning, match="No velocity defined on frame")
    ):
        coord = SpectralCoord([100, 200, 300] * u.nm, observer=observer, target=target)

    if observer is None:
        assert coord.observer is None
    else:
        assert_frame_allclose(observer, coord.observer)

    if target is None:
        assert coord.target is None
    else:
        assert_frame_allclose(target, coord.target)

    assert coord.doppler_rest is None
    assert coord.doppler_convention is None

    if observer is None or target is None:
        assert quantity_allclose(coord.redshift, 0)
        assert quantity_allclose(coord.radial_velocity, 0 * u.km / u.s)
    elif any(observer is lsrd for lsrd in LSRD_EQUIV) and any(
        target is lsrd for lsrd in LSRD_DIR_STATIONARY_EQUIV
    ):
        assert_quantity_allclose(
            coord.radial_velocity, -(274**0.5) * u.km / u.s, atol=1e-4 * u.km / u.s
        )
        assert_quantity_allclose(coord.redshift, -5.5213158163147646e-05, atol=1e-9)
    else:
        raise NotImplementedError()


def test_create_from_spectral_coord(observer, target):
    """
    Checks that parameters are correctly copied to the new SpectralCoord object
    """
    with (
        nullcontext()
        if target is None
        else pytest.warns(AstropyUserWarning, match="No velocity defined on frame")
    ):
        spec_coord1 = SpectralCoord(
            [100, 200, 300] * u.nm,
            observer=observer,
            target=target,
            doppler_convention="optical",
            doppler_rest=6000 * u.AA,
        )
    spec_coord2 = SpectralCoord(spec_coord1)
    assert spec_coord1.observer == spec_coord2.observer
    assert spec_coord1.target == spec_coord2.target
    assert spec_coord1.radial_velocity == spec_coord2.radial_velocity
    assert spec_coord1.doppler_convention == spec_coord2.doppler_convention
    assert spec_coord1.doppler_rest == spec_coord2.doppler_rest


# INTERNAL FUNCTIONS TESTS


def test_apply_relativistic_doppler_shift():
    # Frequency
    sq1 = SpectralQuantity(1 * u.GHz)
    sq2 = _apply_relativistic_doppler_shift(sq1, 0.5 * c)
    assert_quantity_allclose(sq2, np.sqrt(1.0 / 3.0) * u.GHz)

    # Wavelength
    sq3 = SpectralQuantity(500 * u.nm)
    sq4 = _apply_relativistic_doppler_shift(sq3, 0.5 * c)
    assert_quantity_allclose(sq4, np.sqrt(3) * 500 * u.nm)

    # Energy
    sq5 = SpectralQuantity(300 * u.eV)
    sq6 = _apply_relativistic_doppler_shift(sq5, 0.5 * c)
    assert_quantity_allclose(sq6, np.sqrt(1.0 / 3.0) * 300 * u.eV)

    # Wavenumber
    sq7 = SpectralQuantity(0.01 / u.micron)
    sq8 = _apply_relativistic_doppler_shift(sq7, 0.5 * c)
    assert_quantity_allclose(sq8, np.sqrt(1.0 / 3.0) * 0.01 / u.micron)

    # Velocity (doppler_convention='relativistic')
    sq9 = SpectralQuantity(
        200 * u.km / u.s, doppler_convention="relativistic", doppler_rest=1 * u.GHz
    )
    sq10 = _apply_relativistic_doppler_shift(sq9, 300 * u.km / u.s)
    assert_quantity_allclose(sq10, 499.999666 * u.km / u.s)
    assert sq10.doppler_convention == "relativistic"

    # Velocity (doppler_convention='optical')
    sq11 = SpectralQuantity(
        200 * u.km / u.s, doppler_convention="radio", doppler_rest=1 * u.GHz
    )
    sq12 = _apply_relativistic_doppler_shift(sq11, 300 * u.km / u.s)
    assert_quantity_allclose(sq12, 499.650008 * u.km / u.s)
    assert sq12.doppler_convention == "radio"

    # Velocity (doppler_convention='radio')
    sq13 = SpectralQuantity(
        200 * u.km / u.s, doppler_convention="optical", doppler_rest=1 * u.GHz
    )
    sq14 = _apply_relativistic_doppler_shift(sq13, 300 * u.km / u.s)
    assert_quantity_allclose(sq14, 500.350493 * u.km / u.s)
    assert sq14.doppler_convention == "optical"

    # Velocity - check relativistic velocity addition
    sq13 = SpectralQuantity(
        0 * u.km / u.s, doppler_convention="relativistic", doppler_rest=1 * u.GHz
    )
    sq14 = _apply_relativistic_doppler_shift(sq13, 0.999 * c)
    assert_quantity_allclose(sq14, 0.999 * c)
    sq14 = _apply_relativistic_doppler_shift(sq14, 0.999 * c)
    assert_quantity_allclose(sq14, (0.999 * 2) / (1 + 0.999**2) * c)
    assert sq14.doppler_convention == "relativistic"

    # Cases that should raise errors
    sq15 = SpectralQuantity(200 * u.km / u.s)
    with pytest.raises(ValueError, match="doppler_convention not set"):
        _apply_relativistic_doppler_shift(sq15, 300 * u.km / u.s)
    sq16 = SpectralQuantity(200 * u.km / u.s, doppler_rest=10 * u.GHz)
    with pytest.raises(ValueError, match="doppler_convention not set"):
        _apply_relativistic_doppler_shift(sq16, 300 * u.km / u.s)
    sq17 = SpectralQuantity(200 * u.km / u.s, doppler_convention="optical")
    with pytest.raises(ValueError, match="doppler_rest not set"):
        _apply_relativistic_doppler_shift(sq17, 300 * u.km / u.s)


# BASIC TESTS


def test_init_quantity():
    sc = SpectralCoord(10 * u.GHz)
    assert sc.value == 10.0
    assert sc.unit is u.GHz
    assert sc.doppler_convention is None
    assert sc.doppler_rest is None
    assert sc.observer is None
    assert sc.target is None


def test_init_spectral_quantity():
    sc = SpectralCoord(SpectralQuantity(10 * u.GHz, doppler_convention="optical"))
    assert sc.value == 10.0
    assert sc.unit is u.GHz
    assert sc.doppler_convention == "optical"
    assert sc.doppler_rest is None
    assert sc.observer is None
    assert sc.target is None


def test_init_too_many_args():
    with pytest.raises(
        ValueError, match="Cannot specify radial velocity or redshift if both"
    ):
        SpectralCoord(
            10 * u.GHz,
            observer=LSRD,
            target=SkyCoord(10, 20, unit="deg"),
            radial_velocity=1 * u.km / u.s,
        )

    with pytest.raises(
        ValueError, match="Cannot specify radial velocity or redshift if both"
    ):
        SpectralCoord(
            10 * u.GHz, observer=LSRD, target=SkyCoord(10, 20, unit="deg"), redshift=1
        )

    with pytest.raises(
        ValueError, match="Cannot set both a radial velocity and redshift"
    ):
        SpectralCoord(10 * u.GHz, radial_velocity=1 * u.km / u.s, redshift=1)


def test_init_wrong_type():
    with pytest.raises(
        TypeError, match="observer must be a SkyCoord or coordinate frame instance"
    ):
        SpectralCoord(10 * u.GHz, observer=3.4)

    with pytest.raises(
        TypeError, match="target must be a SkyCoord or coordinate frame instance"
    ):
        SpectralCoord(10 * u.GHz, target=3.4)

    with pytest.raises(
        u.UnitsError,
        match=(
            "Argument 'radial_velocity' to function "
            "'__new__' must be in units convertible to 'km / s'"
        ),
    ):
        SpectralCoord(10 * u.GHz, radial_velocity=1 * u.kg)

    with pytest.raises(
        TypeError,
        match=(
            "Argument 'radial_velocity' to function '__new__' has no 'unit' attribute."
            " You should pass in an astropy Quantity instead."
        ),
    ):
        SpectralCoord(10 * u.GHz, radial_velocity="banana")

    with pytest.raises(u.UnitsError, match="redshift should be dimensionless"):
        SpectralCoord(10 * u.GHz, redshift=1 * u.m)

    with pytest.raises(
        TypeError,
        match='Cannot parse "banana" as a Quantity. It does not start with a number.',
    ):
        SpectralCoord(10 * u.GHz, redshift="banana")


def test_observer_init_rv_behavior():
    """
    Test basic initialization behavior or observer/target and redshift/rv
    """

    # Start off by specifying the radial velocity only
    sc_init = SpectralCoord([4000, 5000] * u.AA, radial_velocity=100 * u.km / u.s)
    assert sc_init.observer is None
    assert sc_init.target is None
    assert_quantity_allclose(sc_init.radial_velocity, 100 * u.km / u.s)

    # Next, set the observer, and check that the radial velocity hasn't changed
    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        sc_init.observer = ICRS(CartesianRepresentation([0 * u.km, 0 * u.km, 0 * u.km]))
    assert sc_init.observer is not None
    assert_quantity_allclose(sc_init.radial_velocity, 100 * u.km / u.s)

    # Setting the target should now cause the original radial velocity to be
    # dropped in favor of the automatically computed one
    sc_init.target = SkyCoord(
        CartesianRepresentation([1 * u.km, 0 * u.km, 0 * u.km]),
        frame="icrs",
        radial_velocity=30 * u.km / u.s,
    )
    assert sc_init.target is not None
    assert_quantity_allclose(sc_init.radial_velocity, 30 * u.km / u.s)

    # The observer can only be set if originally None - now that it isn't
    # setting it again should fail
    with pytest.raises(ValueError, match="observer has already been set"):
        sc_init.observer = GCRS(CartesianRepresentation([0 * u.km, 1 * u.km, 0 * u.km]))

    # And similarly, changing the target should not be possible
    with pytest.raises(ValueError, match="target has already been set"):
        sc_init.target = GCRS(CartesianRepresentation([0 * u.km, 1 * u.km, 0 * u.km]))


def test_rv_redshift_initialization():
    # Check that setting the redshift sets the radial velocity appropriately,
    # and that the redshift can be recovered
    sc_init = SpectralCoord([4000, 5000] * u.AA, redshift=1)
    assert isinstance(sc_init.redshift, u.Quantity)
    assert_quantity_allclose(sc_init.redshift, 1 * u.dimensionless_unscaled)
    assert_quantity_allclose(sc_init.radial_velocity, 0.6 * c)

    # Check that setting the same radial velocity produces the same redshift
    # and that the radial velocity can be recovered
    sc_init2 = SpectralCoord([4000, 5000] * u.AA, radial_velocity=0.6 * c)
    assert_quantity_allclose(sc_init2.redshift, 1 * u.dimensionless_unscaled)
    assert_quantity_allclose(sc_init2.radial_velocity, 0.6 * c)

    # Check that specifying redshift as a quantity works
    sc_init3 = SpectralCoord([4000, 5000] * u.AA, redshift=1 * u.one)
    assert sc_init.redshift == sc_init3.redshift

    # Make sure that both redshift and radial velocity can't be specified at
    # the same time.
    with pytest.raises(
        ValueError, match="Cannot set both a radial velocity and redshift"
    ):
        SpectralCoord([4000, 5000] * u.AA, radial_velocity=10 * u.km / u.s, redshift=2)


def test_replicate():
    # The replicate method makes a new object with attributes updated, but doesn't
    # do any conversion

    sc_init = SpectralCoord([4000, 5000] * u.AA, redshift=2)

    sc_set_rv = sc_init.replicate(redshift=1)
    assert_quantity_allclose(sc_set_rv.radial_velocity, 0.6 * c)
    assert_quantity_allclose(sc_init, [4000, 5000] * u.AA)

    sc_set_rv = sc_init.replicate(radial_velocity=c / 2)
    assert_quantity_allclose(sc_set_rv.redshift, np.sqrt(3) - 1)
    assert_quantity_allclose(sc_init, [4000, 5000] * u.AA)

    gcrs_origin = GCRS(CartesianRepresentation([0 * u.km, 0 * u.km, 0 * u.km]))
    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        sc_init2 = SpectralCoord([4000, 5000] * u.AA, redshift=1, observer=gcrs_origin)
    with np.errstate(all="ignore"):
        sc_init2.replicate(redshift=0.5)
    assert_quantity_allclose(sc_init2, [4000, 5000] * u.AA)

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        sc_init3 = SpectralCoord([4000, 5000] * u.AA, redshift=1, target=gcrs_origin)
    with np.errstate(all="ignore"):
        sc_init3.replicate(redshift=0.5)
    assert_quantity_allclose(sc_init2, [4000, 5000] * u.AA)

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        sc_init4 = SpectralCoord(
            [4000, 5000] * u.AA, observer=gcrs_origin, target=gcrs_origin
        )
    with pytest.raises(
        ValueError,
        match=(
            "Cannot specify radial velocity or redshift if both target and observer are"
            " specified"
        ),
    ):
        sc_init4.replicate(redshift=0.5)

    sc_init = SpectralCoord([4000, 5000] * u.AA, redshift=2)
    sc_init_copy = sc_init.replicate(copy=True)
    sc_init[0] = 6000 * u.AA
    assert_quantity_allclose(sc_init_copy, [4000, 5000] * u.AA)

    sc_init = SpectralCoord([4000, 5000] * u.AA, redshift=2)
    sc_init_ref = sc_init.replicate()
    sc_init[0] = 6000 * u.AA
    assert_quantity_allclose(sc_init_ref, [6000, 5000] * u.AA)


def test_with_observer_stationary_relative_to():
    # Simple tests of with_observer_stationary_relative_to to cover different
    # ways of calling it

    # The replicate method makes a new object with attributes updated, but doesn't
    # do any conversion

    sc1 = SpectralCoord([4000, 5000] * u.AA)
    with pytest.raises(
        ValueError,
        match=(
            "This method can only be used if both observer and target are defined on"
            " the SpectralCoord"
        ),
    ):
        sc1.with_observer_stationary_relative_to("icrs")

    sc2 = SpectralCoord(
        [4000, 5000] * u.AA,
        observer=ICRS(
            0 * u.km,
            0 * u.km,
            0 * u.km,
            -1 * u.km / u.s,
            0 * u.km / u.s,
            -1 * u.km / u.s,
            representation_type="cartesian",
            differential_type="cartesian",
        ),
        target=ICRS(
            0 * u.deg, 45 * u.deg, distance=1 * u.kpc, radial_velocity=2 * u.km / u.s
        ),
    )

    # Motion of observer is in opposite direction to target
    assert_quantity_allclose(sc2.radial_velocity, (2 + 2**0.5) * u.km / u.s)

    # Change to observer that is stationary in ICRS
    sc3 = sc2.with_observer_stationary_relative_to("icrs")

    # Velocity difference is now pure radial velocity of target
    assert_quantity_allclose(sc3.radial_velocity, 2 * u.km / u.s)

    # Check setting the velocity in with_observer_stationary_relative_to
    sc4 = sc2.with_observer_stationary_relative_to(
        "icrs", velocity=[-(2**0.5), 0, -(2**0.5)] * u.km / u.s
    )

    # Observer once again moving away from target but faster
    assert_quantity_allclose(sc4.radial_velocity, 4 * u.km / u.s)

    # Check that we can also pass frame classes instead of names

    sc5 = sc2.with_observer_stationary_relative_to(
        ICRS, velocity=[-(2**0.5), 0, -(2**0.5)] * u.km / u.s
    )
    assert_quantity_allclose(sc5.radial_velocity, 4 * u.km / u.s)

    # And make sure we can also pass instances of classes without data

    sc6 = sc2.with_observer_stationary_relative_to(
        ICRS(), velocity=[-(2**0.5), 0, -(2**0.5)] * u.km / u.s
    )
    assert_quantity_allclose(sc6.radial_velocity, 4 * u.km / u.s)

    # And with data provided no velocities are present

    sc7 = sc2.with_observer_stationary_relative_to(
        ICRS(0 * u.km, 0 * u.km, 0 * u.km, representation_type="cartesian"),
        velocity=[-(2**0.5), 0, -(2**0.5)] * u.km / u.s,
    )
    assert_quantity_allclose(sc7.radial_velocity, 4 * u.km / u.s)

    # And also have the ability to pass frames with velocities already defined

    sc8 = sc2.with_observer_stationary_relative_to(
        ICRS(
            0 * u.km,
            0 * u.km,
            0 * u.km,
            2**0.5 * u.km / u.s,
            0 * u.km / u.s,
            2**0.5 * u.km / u.s,
            representation_type="cartesian",
            differential_type="cartesian",
        )
    )
    assert_quantity_allclose(
        sc8.radial_velocity, 0 * u.km / u.s, atol=1e-10 * u.km / u.s
    )

    # Make sure that things work properly if passing a SkyCoord

    sc9 = sc2.with_observer_stationary_relative_to(
        SkyCoord(ICRS(0 * u.km, 0 * u.km, 0 * u.km, representation_type="cartesian")),
        velocity=[-(2**0.5), 0, -(2**0.5)] * u.km / u.s,
    )
    assert_quantity_allclose(sc9.radial_velocity, 4 * u.km / u.s)

    sc10 = sc2.with_observer_stationary_relative_to(
        SkyCoord(
            ICRS(
                0 * u.km,
                0 * u.km,
                0 * u.km,
                2**0.5 * u.km / u.s,
                0 * u.km / u.s,
                2**0.5 * u.km / u.s,
                representation_type="cartesian",
                differential_type="cartesian",
            )
        )
    )
    assert_quantity_allclose(
        sc10.radial_velocity, 0 * u.km / u.s, atol=1e-10 * u.km / u.s
    )

    # But we shouldn't be able to pass both a frame with velocities, and explicit velocities

    with pytest.raises(
        ValueError,
        match="frame already has differentials, cannot also specify velocity",
    ):
        sc2.with_observer_stationary_relative_to(
            ICRS(
                0 * u.km,
                0 * u.km,
                0 * u.km,
                2**0.5 * u.km / u.s,
                0 * u.km / u.s,
                2**0.5 * u.km / u.s,
                representation_type="cartesian",
                differential_type="cartesian",
            ),
            velocity=[-(2**0.5), 0, -(2**0.5)] * u.km / u.s,
        )

    # And velocities should have three elements

    with pytest.raises(
        ValueError, match="velocity should be a Quantity vector with 3 elements"
    ):
        sc2.with_observer_stationary_relative_to(
            ICRS, velocity=[-(2**0.5), 0, -(2**0.5), -3] * u.km / u.s
        )

    # Make sure things don't change depending on what frame class is used for reference
    sc11 = sc2.with_observer_stationary_relative_to(
        SkyCoord(
            ICRS(
                0 * u.km,
                0 * u.km,
                0 * u.km,
                2**0.5 * u.km / u.s,
                0 * u.km / u.s,
                2**0.5 * u.km / u.s,
                representation_type="cartesian",
                differential_type="cartesian",
            )
        ).transform_to(Galactic)
    )
    assert_quantity_allclose(
        sc11.radial_velocity, 0 * u.km / u.s, atol=1e-10 * u.km / u.s
    )

    # Check that it is possible to preserve the observer frame
    sc12 = sc2.with_observer_stationary_relative_to(LSRD)
    sc13 = sc2.with_observer_stationary_relative_to(LSRD, preserve_observer_frame=True)

    assert isinstance(sc12.observer, Galactic)
    assert isinstance(sc13.observer, ICRS)


def test_los_shift_radial_velocity():
    # Tests to make sure that with_radial_velocity_shift correctly calculates
    # the new radial velocity

    # First check case where observer and/or target aren't specified

    sc1 = SpectralCoord(500 * u.nm, radial_velocity=1 * u.km / u.s)

    sc2 = sc1.with_radial_velocity_shift(1 * u.km / u.s)
    assert_quantity_allclose(sc2.radial_velocity, 2 * u.km / u.s)

    sc3 = sc1.with_radial_velocity_shift(-3 * u.km / u.s)
    assert_quantity_allclose(sc3.radial_velocity, -2 * u.km / u.s)

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        sc4 = SpectralCoord(
            500 * u.nm, radial_velocity=1 * u.km / u.s, observer=gcrs_not_origin
        )

    sc5 = sc4.with_radial_velocity_shift(1 * u.km / u.s)
    assert_quantity_allclose(sc5.radial_velocity, 2 * u.km / u.s)

    sc6 = sc4.with_radial_velocity_shift(-3 * u.km / u.s)
    assert_quantity_allclose(sc6.radial_velocity, -2 * u.km / u.s)

    if PYTEST_LT_8_0:
        ctx = nullcontext()
    else:
        ctx = pytest.warns(
            NoDistanceWarning, match="Distance on coordinate object is dimensionless"
        )

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"), ctx:
        sc7 = SpectralCoord(
            500 * u.nm,
            radial_velocity=1 * u.km / u.s,
            target=ICRS(10 * u.deg, 20 * u.deg),
        )

    sc8 = sc7.with_radial_velocity_shift(1 * u.km / u.s)
    assert_quantity_allclose(sc8.radial_velocity, 2 * u.km / u.s)

    sc9 = sc7.with_radial_velocity_shift(-3 * u.km / u.s)
    assert_quantity_allclose(sc9.radial_velocity, -2 * u.km / u.s)

    # Check that things still work when both observer and target are specified

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        sc10 = SpectralCoord(
            500 * u.nm,
            observer=ICRS(0 * u.deg, 0 * u.deg, distance=1 * u.m),
            target=ICRS(
                10 * u.deg,
                20 * u.deg,
                radial_velocity=1 * u.km / u.s,
                distance=10 * u.kpc,
            ),
        )

    sc11 = sc10.with_radial_velocity_shift(1 * u.km / u.s)
    assert_quantity_allclose(sc11.radial_velocity, 2 * u.km / u.s)

    sc12 = sc10.with_radial_velocity_shift(-3 * u.km / u.s)
    assert_quantity_allclose(sc12.radial_velocity, -2 * u.km / u.s)

    # Check that things work if radial_velocity wasn't specified at all

    sc13 = SpectralCoord(500 * u.nm)
    sc14 = sc13.with_radial_velocity_shift(1 * u.km / u.s)
    assert_quantity_allclose(sc14.radial_velocity, 1 * u.km / u.s)

    sc15 = sc1.with_radial_velocity_shift()
    assert_quantity_allclose(sc15.radial_velocity, 1 * u.km / u.s)

    # Check that units are verified

    with pytest.raises(
        u.UnitsError,
        match=(
            "Argument must have unit physical type 'speed' for radial velocty or "
            "'dimensionless' for redshift."
        ),
    ):
        sc1.with_radial_velocity_shift(target_shift=1 * u.kg)


@pytest.mark.xfail
def test_relativistic_radial_velocity():
    # Test for when both observer and target have relativistic velocities.
    # This is not yet supported, so the test is xfailed for now.

    sc = SpectralCoord(
        500 * u.nm,
        observer=ICRS(
            0 * u.km,
            0 * u.km,
            0 * u.km,
            -0.5 * c,
            -0.5 * c,
            -0.5 * c,
            representation_type="cartesian",
            differential_type="cartesian",
        ),
        target=ICRS(
            1 * u.kpc,
            1 * u.kpc,
            1 * u.kpc,
            0.5 * c,
            0.5 * c,
            0.5 * c,
            representation_type="cartesian",
            differential_type="cartesian",
        ),
    )

    assert_quantity_allclose(sc.radial_velocity, 0.989743318610787 * u.km / u.s)


# SCIENCE USE CASE TESTS


def test_spectral_coord_jupiter():
    """
    Checks radial velocity between Earth and Jupiter
    """
    obstime = time.Time("2018-12-13 9:00")
    obs = GREENWICH.get_gcrs(obstime)

    pos, vel = get_body_barycentric_posvel("jupiter", obstime)
    jupiter = SkyCoord(
        pos.with_differentials(CartesianDifferential(vel.xyz)), obstime=obstime
    )

    spc = SpectralCoord([100, 200, 300] * u.nm, observer=obs, target=jupiter)

    # The velocity should be less than ~43 + a bit extra, which is the
    # maximum possible earth-jupiter relative velocity. We check the exact
    # value here (determined from SpectralCoord, so this serves as a test to
    # check that this value doesn't change - the value is not a ground truth)
    assert_quantity_allclose(spc.radial_velocity, -7.35219854 * u.km / u.s)


def test_spectral_coord_alphacen():
    """
    Checks radial velocity between Earth and Alpha Centauri
    """
    obstime = time.Time("2018-12-13 9:00")
    obs = GREENWICH.get_gcrs(obstime)

    # Coordinates were obtained from the following then hard-coded to avoid download
    # acen = SkyCoord.from_name('alpha cen')
    acen = SkyCoord(
        ra=219.90085 * u.deg,
        dec=-60.83562 * u.deg,
        frame="icrs",
        distance=4.37 * u.lightyear,
        radial_velocity=-18.0 * u.km / u.s,
    )

    spc = SpectralCoord([100, 200, 300] * u.nm, observer=obs, target=acen)

    # The velocity should be less than ~18 + 30 + a bit extra, which is the
    # maximum possible relative velocity.  We check the exact value here
    # (determined from SpectralCoord, so this serves as a test to check that
    # this value doesn't change - the value is not a ground truth)
    assert_quantity_allclose(spc.radial_velocity, -26.328301 * u.km / u.s)


def test_spectral_coord_m31():
    """
    Checks radial velocity between Earth and M31
    """
    obstime = time.Time("2018-12-13 9:00")
    obs = GREENWICH.get_gcrs(obstime)

    # Coordinates were obtained from the following then hard-coded to avoid download
    # m31 = SkyCoord.from_name('M31')
    m31 = SkyCoord(
        ra=10.6847 * u.deg,
        dec=41.269 * u.deg,
        distance=710 * u.kpc,
        radial_velocity=-300 * u.km / u.s,
    )

    spc = SpectralCoord([100, 200, 300] * u.nm, observer=obs, target=m31)

    # The velocity should be less than ~300 + 30 + a bit extra in km/s, which
    # is the maximum possible relative velocity.  We check the exact values
    # here (determined from SpectralCoord, so this serves as a test to check
    # that this value doesn't change - the value is not a ground truth)
    assert_quantity_allclose(spc.radial_velocity, -279.755128 * u.km / u.s)
    assert_allclose(spc.redshift, -0.0009327276702120191)


def test_shift_to_rest_galaxy():
    """
    This tests storing a spectral coordinate with a specific redshift, and then
    doing basic rest-to-observed-and-back transformations
    """
    z = 5
    rest_line_wls = [5007, 6563] * u.AA

    observed_spc = SpectralCoord(rest_line_wls * (z + 1), redshift=z)
    rest_spc = observed_spc.to_rest()
    # alternatively:
    # rest_spc = observed_spc.with_observer(observed_spec.target)
    # although then it would have to be clearly documented, or the `to_rest`
    # implemented in Spectrum1D?

    assert_quantity_allclose(rest_spc, rest_line_wls)

    # No frames are explicitly defined, so to the user, the observer and
    #  target are not set.
    with pytest.raises(AttributeError):
        assert_frame_allclose(rest_spc.observer, rest_spc.target)


def test_shift_to_rest_star_withobserver():
    rv = -8.3283011 * u.km / u.s
    rest_line_wls = [5007, 6563] * u.AA

    obstime = time.Time("2018-12-13 9:00")
    obs = GREENWICH.get_gcrs(obstime)
    acen = SkyCoord(
        ra=219.90085 * u.deg,
        dec=-60.83562 * u.deg,
        frame="icrs",
        distance=4.37 * u.lightyear,
    )
    # Note that above the rv is missing from the SkyCoord.
    # That's intended, as it will instead be set in the `SpectralCoord`.  But
    # the SpectralCoord machinery should yield something comparable to test_
    # spectral_coord_alphacen

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        observed_spc = SpectralCoord(
            rest_line_wls * (rv / c + 1), observer=obs, target=acen
        )

    rest_spc = observed_spc.to_rest()
    assert_quantity_allclose(rest_spc, rest_line_wls)

    barycentric_spc = observed_spc.with_observer_stationary_relative_to("icrs")
    baryrest_spc = barycentric_spc.to_rest()
    assert quantity_allclose(baryrest_spc, rest_line_wls)

    # now make sure the change the barycentric shift did is comparable to the
    # offset rv_correction produces
    # barytarg = SkyCoord(barycentric_spc.target.frame)  # should be this but that doesn't work for unclear reasons
    barytarg = SkyCoord(
        barycentric_spc.target.data.without_differentials(),
        frame=barycentric_spc.target.realize_frame(None),
    )
    vcorr = barytarg.radial_velocity_correction(
        kind="barycentric", obstime=obstime, location=GREENWICH
    )

    drv = baryrest_spc.radial_velocity - observed_spc.radial_velocity

    # note this probably will not work on the first try, but it's ok if this is
    # "good enough", where good enough is estimated below.  But that could be
    # adjusted if we think that's too aggressive of a precision target for what
    # the machinery can handle
    # with pytest.raises(AssertionError):
    assert_quantity_allclose(vcorr, drv, atol=10 * u.m / u.s)


gcrs_origin = GCRS(CartesianRepresentation([0 * u.km, 0 * u.km, 0 * u.km]))
gcrs_not_origin = GCRS(CartesianRepresentation([1 * u.km, 0 * u.km, 0 * u.km]))


@pytest.mark.parametrize(
    "sc_kwargs",
    [
        {"radial_velocity": 0 * u.km / u.s},
        {"observer": gcrs_origin, "radial_velocity": 0 * u.km / u.s},
        {"target": gcrs_origin, "radial_velocity": 0 * u.km / u.s},
        {"observer": gcrs_origin, "target": gcrs_not_origin},
    ],
)
def test_los_shift(sc_kwargs):
    wl = [4000, 5000] * u.AA
    with (
        nullcontext()
        if "observer" not in sc_kwargs and "target" not in sc_kwargs
        else pytest.warns(AstropyUserWarning, match="No velocity defined on frame")
    ):
        sc_init = SpectralCoord(wl, **sc_kwargs)

    # these should always work in *all* cases because it's unambiguous that
    # a target shift should behave this way
    new_sc1 = sc_init.with_radial_velocity_shift(0.1)
    assert_quantity_allclose(new_sc1, wl * 1.1)
    # interpret at redshift
    new_sc2 = sc_init.with_radial_velocity_shift(0.1 * u.dimensionless_unscaled)
    assert_quantity_allclose(new_sc1, new_sc2)

    new_sc3 = sc_init.with_radial_velocity_shift(-100 * u.km / u.s)
    assert_quantity_allclose(new_sc3, wl * (1 + (-100 * u.km / u.s / c)))

    # now try the cases where observer is specified as well/instead
    if sc_init.observer is None or sc_init.target is None:
        with pytest.raises(ValueError):
            # both must be specified if you're going to mess with observer
            sc_init.with_radial_velocity_shift(observer_shift=0.1)

    if sc_init.observer is not None and sc_init.target is not None:
        # redshifting the observer should *blushift* the LOS velocity since
        # its the observer-to-target vector that matters
        new_sc4 = sc_init.with_radial_velocity_shift(observer_shift=0.1)
        assert_quantity_allclose(new_sc4, wl / 1.1)

        # an equal shift in both should produce no offset at all
        new_sc5 = sc_init.with_radial_velocity_shift(
            target_shift=0.1, observer_shift=0.1
        )
        assert_quantity_allclose(new_sc5, wl)


def test_asteroid_velocity_frame_shifts():
    """
    This test mocks up the use case of observing a spectrum of an asteroid
    at different times and from different observer locations.
    """
    time1 = time.Time("2018-12-13 9:00")
    dt = 12 * u.hour
    time2 = time1 + dt

    # make the silly but simplifying assumption that the asteroid is moving along
    # the x-axis of GCRS, and makes a 10 earth-radius closest approach

    v_ast = [5, 0, 0] * u.km / u.s
    x1 = -v_ast[0] * dt / 2
    x2 = v_ast[0] * dt / 2
    z = 10 * u.Rearth

    cdiff = CartesianDifferential(v_ast)

    asteroid_loc1 = GCRS(
        CartesianRepresentation(x1.to(u.km), 0 * u.km, z.to(u.km), differentials=cdiff),
        obstime=time1,
    )
    asteroid_loc2 = GCRS(
        CartesianRepresentation(x2.to(u.km), 0 * u.km, z.to(u.km), differentials=cdiff),
        obstime=time2,
    )

    # assume satellites that are essentially fixed in geostationary orbit on
    # opposite sides of the earth
    observer1 = GCRS(
        CartesianRepresentation([0 * u.km, 35000 * u.km, 0 * u.km]), obstime=time1
    )
    observer2 = GCRS(
        CartesianRepresentation([0 * u.km, -35000 * u.km, 0 * u.km]), obstime=time2
    )

    wls = np.linspace(4000, 7000, 100) * u.AA
    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        spec_coord1 = SpectralCoord(wls, observer=observer1, target=asteroid_loc1)

    assert spec_coord1.radial_velocity < 0 * u.km / u.s
    assert spec_coord1.radial_velocity > -5 * u.km / u.s

    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        spec_coord2 = SpectralCoord(wls, observer=observer2, target=asteroid_loc2)

    assert spec_coord2.radial_velocity > 0 * u.km / u.s
    assert spec_coord2.radial_velocity < 5 * u.km / u.s

    # now check the behavior of with_observer_stationary_relative_to: we shift each coord
    # into the velocity frame of its *own* target.  That would then be a
    # spectralcoord that would allow direct physical comparison of the two
    # different spec_corrds.  There's no way to test that, without
    # actual data, though.

    # spec_coord2 is redshifted, so we test that it behaves the way "shifting
    # to rest frame" should - the as-observed spectral coordinate should become
    # the rest frame, so something that starts out red should become bluer
    target_sc2 = spec_coord2.with_observer_stationary_relative_to(spec_coord2.target)
    assert np.all(target_sc2 < spec_coord2)
    # rv/redshift should be 0 since the observer and target velocities should
    # be the same
    assert_quantity_allclose(
        target_sc2.radial_velocity, 0 * u.km / u.s, atol=1e-7 * u.km / u.s
    )

    # check that the same holds for spec_coord1, but be more specific: it
    # should follow the standard redshift formula (which in this case yields
    # a blueshift, although the formula is the same as 1+z)
    target_sc1 = spec_coord1.with_observer_stationary_relative_to(spec_coord1.target)
    assert_quantity_allclose(target_sc1, spec_coord1 / (1 + spec_coord1.redshift))

    # TODO: Figure out what is meant by the below use case
    # ensure the "target-rest" use gives the same answer
    # target_sc1_alt = spec_coord1.with_observer_stationary_relative_to('target-rest')
    # assert_quantity_allclose(target_sc1, target_sc1_alt)


def test_spectral_coord_from_sky_coord_without_distance():
    # see https://github.com/astropy/specutils/issues/658 for issue context
    obs = SkyCoord(0 * u.m, 0 * u.m, 0 * u.m, representation_type="cartesian")
    with pytest.warns(AstropyUserWarning, match="No velocity defined on frame"):
        coord = SpectralCoord([1, 2, 3] * u.micron, observer=obs)
    # coord.target = SkyCoord.from_name('m31')  # <- original issue, but below is the same but requires no remote data access
    if PYTEST_LT_8_0:
        ctx = nullcontext()
    else:
        ctx = pytest.warns(NoVelocityWarning, match="No velocity defined on frame")
    with (
        pytest.warns(
            AstropyUserWarning, match="Distance on coordinate object is dimensionless"
        ),
        ctx,
    ):
        coord.target = SkyCoord(ra=10.68470833 * u.deg, dec=41.26875 * u.deg)


EXPECTED_VELOCITY_FRAMES = {
    "geocent": "gcrs",
    "heliocent": "hcrs",
    "lsrk": "lsrk",
    "lsrd": "lsrd",
    "galactoc": FITSWCS_VELOCITY_FRAMES["GALACTOC"],
    "localgrp": FITSWCS_VELOCITY_FRAMES["LOCALGRP"],
}


@pytest.mark.parametrize("specsys", list(EXPECTED_VELOCITY_FRAMES))
@pytest.mark.slow
def test_spectralcoord_accuracy(specsys):
    # This is a test to check the numerical results of transformations between
    # different velocity frames in SpectralCoord. This compares the velocity
    # shifts determined with SpectralCoord to those determined from the rv
    # package in Starlink.

    velocity_frame = EXPECTED_VELOCITY_FRAMES[specsys]

    reference_filename = get_pkg_data_filename("accuracy/data/rv.ecsv")
    reference_table = Table.read(reference_filename, format="ascii.ecsv")

    rest = 550 * u.nm

    if PYTEST_LT_8_0:
        ctx = nullcontext()
    else:
        ctx = pytest.warns(
            NoVelocityWarning,
            match=(
                r"^No velocity defined on frame, assuming \(0\., 0\., 0\.\) km / s\.$"
            ),
        )
    with iers.conf.set_temp("auto_download", False):
        for row in reference_table:
            observer = EarthLocation.from_geodetic(
                -row["obslon"], row["obslat"]
            ).get_itrs(obstime=row["obstime"])

            with (
                ctx,
                pytest.warns(
                    NoDistanceWarning,
                    match=(
                        "^Distance on coordinate object is dimensionless, an arbitrary "
                        r"distance value of 1000000\.0 kpc will be set instead\.$"
                    ),
                ),
            ):
                sc_topo = SpectralCoord(
                    545 * u.nm, observer=observer, target=row["target"]
                )

            # FIXME: A warning is emitted for dates after MJD=57754.0 even
            # though the leap second table should be valid until the end of
            # 2020.
            with (
                nullcontext()
                if row["obstime"].mjd < 57754
                else pytest.warns(AstropyWarning, match="Tried to get polar motions")
            ):
                sc_final = sc_topo.with_observer_stationary_relative_to(velocity_frame)

            delta_vel = sc_topo.to(
                u.km / u.s, doppler_convention="relativistic", doppler_rest=rest
            ) - sc_final.to(
                u.km / u.s, doppler_convention="relativistic", doppler_rest=rest
            )

            if specsys == "galactoc":
                assert_allclose(
                    delta_vel.to_value(u.km / u.s), row[specsys.lower()], atol=30
                )
            else:
                assert_allclose(
                    delta_vel.to_value(u.km / u.s),
                    row[specsys.lower()],
                    atol=0.02,
                    rtol=0.002,
                )


# TODO: add test when target is not ICRS
# TODO: add test when SpectralCoord is in velocity to start with



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
All of the pytest fixtures used by astropy.table are defined here.

`conftest.py` is a "special" module name for pytest that is always
imported, but is not looked in for tests, and it is the recommended
place to put fixtures that are shared between modules.  These fixtures
can not be defined in a module by a different name and still be shared
between modules.
"""

import pickle
from collections import OrderedDict
from copy import deepcopy

import numpy as np
import pytest

from astropy import coordinates, table, time
from astropy import units as u
from astropy.table import QTable, Table, pprint
from astropy.table.table_helpers import ArrayWrapper


@pytest.fixture(params=[table.Column, table.MaskedColumn])
def Column(request):
    # Fixture to run all the Column tests for both an unmasked (ndarray)
    # and masked (MaskedArray) column.
    return request.param


class MaskedTable(table.Table):
    def __init__(self, *args, **kwargs):
        kwargs["masked"] = True
        table.Table.__init__(self, *args, **kwargs)


class MyRow(table.Row):
    pass


class MyColumn(table.Column):
    pass


class MyMaskedColumn(table.MaskedColumn):
    pass


class MyTableColumns(table.TableColumns):
    pass


class MyTableFormatter(pprint.TableFormatter):
    pass


class MyTable(table.Table):
    Row = MyRow
    Column = MyColumn
    MaskedColumn = MyMaskedColumn
    TableColumns = MyTableColumns
    TableFormatter = MyTableFormatter


# Fixture to run all the Column tests for both an unmasked (ndarray)
# and masked (MaskedArray) column.


@pytest.fixture(params=["unmasked", "masked", "subclass"])
def table_types(request):
    class TableTypes:
        def __init__(self, request):
            if request.param == "unmasked":
                self.Table = table.Table
                self.Column = table.Column
            elif request.param == "masked":
                self.Table = MaskedTable
                self.Column = table.MaskedColumn
            elif request.param == "subclass":
                self.Table = MyTable
                self.Column = MyColumn

    return TableTypes(request)


# Fixture to run all the Column tests for both an unmasked (ndarray)
# and masked (MaskedArray) column.
@pytest.fixture(params=[False, True])
def table_data(request):
    class TableData:
        def __init__(self, request):
            self.Table = MaskedTable if request.param else table.Table
            self.Column = table.MaskedColumn if request.param else table.Column
            self.COLS = [
                self.Column(
                    name="a",
                    data=[1, 2, 3],
                    description="da",
                    format="%i",
                    meta={"ma": 1},
                    unit="ua",
                ),
                self.Column(
                    name="b",
                    data=[4, 5, 6],
                    description="db",
                    format="%d",
                    meta={"mb": 1},
                    unit="ub",
                ),
                self.Column(
                    name="c",
                    data=[7, 8, 9],
                    description="dc",
                    format="%f",
                    meta={"mc": 1},
                    unit="ub",
                ),
            ]
            self.DATA = self.Table(self.COLS)

    return TableData(request)


class SubclassTable(table.Table):
    pass


@pytest.fixture(params=[True, False])
def tableclass(request):
    return table.Table if request.param else SubclassTable


@pytest.fixture(params=list(range(pickle.HIGHEST_PROTOCOL + 1)))
def protocol(request):
    """
    Fixture to run all the tests for all available pickle protocols.
    """
    return request.param


# Fixture to run all tests for both an unmasked (ndarray) and masked
# (MaskedArray) column.
@pytest.fixture(params=[False, True])
def table_type(request):
    return MaskedTable if request.param else table.Table


# Stuff for testing mixin columns

MIXIN_COLS = {
    "quantity": [0, 1, 2, 3] * u.m,
    "longitude": coordinates.Longitude(
        [0.0, 1.0, 5.0, 6.0] * u.deg, wrap_angle=180.0 * u.deg
    ),
    "latitude": coordinates.Latitude([5.0, 6.0, 10.0, 11.0] * u.deg),
    "time": time.Time([2000, 2001, 2002, 2003], format="jyear"),
    "timedelta": time.TimeDelta([1, 2, 3, 4], format="jd"),
    "skycoord": coordinates.SkyCoord(ra=[0, 1, 2, 3] * u.deg, dec=[0, 1, 2, 3] * u.deg),
    "sphericalrep": coordinates.SphericalRepresentation(
        [0, 1, 2, 3] * u.deg, [0, 1, 2, 3] * u.deg, 1 * u.kpc
    ),
    "cartesianrep": coordinates.CartesianRepresentation(
        [0, 1, 2, 3] * u.pc, [4, 5, 6, 7] * u.pc, [9, 8, 8, 6] * u.pc
    ),
    "sphericaldiff": coordinates.SphericalCosLatDifferential(
        [0, 1, 2, 3] * u.mas / u.yr, [0, 1, 2, 3] * u.mas / u.yr, 10 * u.km / u.s
    ),
    "arraywrap": ArrayWrapper([0, 1, 2, 3]),
    "arrayswap": ArrayWrapper(
        np.arange(4, dtype="i").byteswap().view(np.dtype("i").newbyteorder())
    ),
    "ndarraylil": np.array(
        [(7, "a"), (8, "b"), (9, "c"), (9, "c")], dtype="<i4,|S1"
    ).view(table.NdarrayMixin),
    "ndarraybig": np.array(
        [(7, "a"), (8, "b"), (9, "c"), (9, "c")], dtype=">i4,|S1"
    ).view(table.NdarrayMixin),
    "stokescoord": coordinates.StokesCoord(range(1, 5)),
}
MIXIN_COLS["earthlocation"] = coordinates.EarthLocation(
    lon=MIXIN_COLS["longitude"],
    lat=MIXIN_COLS["latitude"],
    height=MIXIN_COLS["quantity"],
)
MIXIN_COLS["sphericalrepdiff"] = coordinates.SphericalRepresentation(
    MIXIN_COLS["sphericalrep"], differentials=MIXIN_COLS["sphericaldiff"]
)


@pytest.fixture(params=sorted(MIXIN_COLS))
def mixin_cols(request):
    """
    Fixture to return a set of columns for mixin testing which includes
    an index column 'i', two string cols 'a', 'b' (for joins etc), and
    one of the available mixin column types.
    """
    cols = OrderedDict()
    mixin_cols = deepcopy(MIXIN_COLS)
    cols["i"] = table.Column([0, 1, 2, 3], name="i")
    cols["a"] = table.Column(["a", "b", "b", "c"], name="a")
    cols["b"] = table.Column(["b", "c", "a", "d"], name="b")
    cols["m"] = mixin_cols[request.param]

    return cols


def _get_test_table():
    T = QTable.read(
        [
            " a b c d",
            " 2 c 7.0 0",
            " 2 b 5.0 1",
            " 2 b 6.0 2",
            " 2 a 4.0 3",
            " 0 a 0.0 4",
            " 1 b 3.0 5",
            " 1 a 2.0 6",
            " 1 a 1.0 7",
        ],
        format="ascii",
    )
    T["q"] = np.arange(len(T)) * u.m
    T.meta.update({"ta": 1})
    T["c"].meta.update({"a": 1})
    T["c"].description = "column c"
    return T


@pytest.fixture()
def T1b(request):
    """Basic table"""
    T = _get_test_table()
    return T


@pytest.fixture(params=[False, True])
def T1(request):
    """Basic table with or without index on integer column a"""
    T = _get_test_table()
    if request.param:
        T.add_index("a")
    return T


@pytest.fixture(params=[False, True])
def T1q(request):
    """Basic table where a column is integer or Quantity"""
    T = _get_test_table()
    if request.param:
        T["a"] = T["a"] * u.m
    return T


@pytest.fixture(params=[(False, False), (False, True), (True, False), (True, True)])
def T1m(request):
    """Basic table with or without index on column a, where a is integer or Quantity"""
    T = _get_test_table()
    add_index, is_quantity = request.param
    if is_quantity:
        T["a"] = T["a"] * u.m
    if add_index:
        T.add_index("a")
    return T


@pytest.fixture(params=[Table, QTable])
def operation_table_type(request):
    return request.param
